# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the functions needed for computing integrals over basis functions.
"""
# pylint: disable= unbalanced-tuple-unpacking, too-many-arguments
import itertools as it

import autograd.numpy as anp
import autograd.scipy as asp
from scipy.special import factorial2 as fac2


def primitive_norm(l, alpha):
    r"""Compute the normalization constant for a primitive Gaussian function.

    A Gaussian function centred at the position :math:`r = (x, y, z)` is defined as

    .. math::

        G = x^{l_x} y^{l_y} z^{l_z} e^{-\alpha r^2},

    where :math:`l = (l_x, l_y, l_z)` defines the angular momentum quantum number and :math:`\alpha`
    is the Gaussian function exponent. The normalization constant for this function is computed as

    .. math::

        N(l, \alpha) = (\frac{2\alpha}{\pi})^{3/4} \frac{(4 \alpha)^{(l_x + l_y + l_z)/2}}
        {(2l_x-1)!! (2l_y-1)!! (2l_z-1)!!)^{1/2}}.

    Args:
        l (tuple[int]): angular momentum quantum number of the basis function
        alpha (array[float]): exponent of the primitive Gaussian function

    Returns:
        array[float]: normalization coefficient

    **Example**

    >>> l = (0, 0, 0)
    >>> alpha = np.array([3.425250914])
    >>> n = primitive_norm(l, alpha)
    >>> print(n)
    array([1.79444183])
    """
    lx, ly, lz = l
    n = (
        (2 * alpha / anp.pi) ** 0.75
        * (4 * alpha) ** (sum(l) / 2)
        / anp.sqrt(fac2(2 * lx - 1) * fac2(2 * ly - 1) * fac2(2 * lz - 1))
    )
    return n


def contracted_norm(l, alpha, a):
    r"""Compute the normalization constant for a contracted Gaussian function.

    A contracted Gaussian function is defined as

    .. math::

        \psi = a_1 G_1 + a_2 G_2 + a_3 G_3,

    where :math:`a` denotes the contraction coefficients and :math:`G` is a primitive Gaussian function. The
    normalization constant for this function is computed as

    .. math::

        N(l, \alpha, a) = [\frac{\pi^{3/2}(2l_x-1)!! (2l_y-1)!! (2l_z-1)!!}{2^{l_x + l_y + l_z}}
        \sum_{i,j} \frac{a_i a_j}{(\alpha_i + \alpha_j)^{{l_x + l_y + l_z+3/2}}}]^{-1/2}

    where :math:`l` and :math:`\alpha` denote the angular momentum quantum number and the exponent
    of the Gaussian function, respectively.

    Args:
        l (tuple[int]): angular momentum quantum number of the primitive Gaussian functions
        alpha (array[float]): exponents of the primitive Gaussian functions
        a (array[float]): coefficients of the contracted Gaussian functions

    Returns:
        array[float]: normalization coefficient

    **Example**

    >>> l = (0, 0, 0)
    >>> alpha = np.array([3.425250914, 0.6239137298, 0.168855404])
    >>> a = np.array([1.79444183, 0.50032649, 0.18773546])
    >>> n = contracted_norm(l, alpha, a)
    >>> print(n)
    0.39969026908800853
    """
    lx, ly, lz = l
    c = anp.pi**1.5 / 2 ** sum(l) * fac2(2 * lx - 1) * fac2(2 * ly - 1) * fac2(2 * lz - 1)
    s = (
        (a.reshape(len(a), 1) * a) / ((alpha.reshape(len(alpha), 1) + alpha) ** (sum(l) + 1.5))
    ).sum()
    n = 1 / anp.sqrt(c * s)
    return n


def _generate_params(params, args):
    """Generate basis set parameters. The default values are used for the non-differentiable
    parameters and the user-defined values are used for the differentiable ones.

    Args:
        params (list(array[float])): default values of the basis set parameters
        args (list(array[float])): initial values of the differentiable basis set parameters

    Returns:
        list(array[float]): basis set parameters
    """
    basis_params = []
    c = 0
    for p in params:
        if p.requires_grad:
            basis_params.append(args[c])
            c += 1
        else:
            basis_params.append(p)
    return basis_params


def expansion(la, lb, ra, rb, alpha, beta, t):
    r"""Compute Hermite Gaussian expansion coefficients recursively for two Gaussian functions.

    An overlap distribution, which defines the product of two Gaussians, can be written as a Hermite
    expansion as [`Helgaker (1995) p798 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]

    .. math::

        \Omega_{ij} = \sum_{t=0}^{i+j} E_t^{ij} \Lambda_t,

    where :math:`\Lambda` is a Hermite polynomial of degree :math:`t`, :math:`E` denotes the expansion
    coefficients, :math:`\Omega_{ij} = G_i G_j`, and :math:`G` is a Gaussian function. The overlap
    integral between two Gaussian functions can be simply computed by integrating over the overlap
    distribution which requires obtaining the expansion coefficients. This can be done recursively
    as [`Helgaker (1995) p799 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]

    .. math::

        E_t^{i+1,j} = \frac{1}{2p} E_{t-1}^{ij} - \frac{qr}{\alpha} E_{t}^{ij} + (t+1) E_{t+1}^{ij},

    and

    .. math::

        E_t^{i,j+1} = \frac{1}{2p} E_{t-1}^{ij} + \frac{qr}{\beta} E_{t}^{ij} + (t+1) E_{t+1}^{ij},

    where :math:`p = \alpha + \beta` and :math:`q = \alpha \beta / (\alpha + \beta)` are computed
    from the Gaussian exponents :math:`\alpha, \beta` and the position :math:`r` is computed as
    :math:`r = r_\alpha - r_\beta`. The starting coefficient is

    .. math::

        E_0^{00} = e^{-qr^2},

    and :math:`E_t^{ij} = 0` if :math:`t < 0` or :math:`t > (i+j)`.

    Args:
        la (integer): angular momentum component for the first Gaussian function
        lb (integer): angular momentum component for the second Gaussian function
        ra (float): position component of the the first Gaussian function
        rb (float): position component of the the second Gaussian function
        alpha (array[float]): exponent of the first Gaussian function
        beta (array[float]): exponent of the second Gaussian function
        t (integer): number of nodes in the Hermite Gaussian

    Returns:
        array[float]: expansion coefficients for each Gaussian combination

    **Example**

    >>> la, lb = 0, 0
    >>> ra, rb = 0.0, 0.0
    >>> alpha = np.array([3.42525091])
    >>> beta =  np.array([3.42525091])
    >>> t = 0
    >>> c = expansion(la, lb, ra, rb, alpha, beta, t)
    >>> c
    array([1.])
    """
    p = anp.array(alpha + beta)
    q = anp.array(alpha * beta / p)
    r = anp.array(ra - rb)

    if la == lb == t == 0:
        return anp.exp(-q * r**2)

    if t < 0 or t > (la + lb):
        return 0.0

    if lb == 0:
        return (
            (1 / (2 * p)) * expansion(la - 1, lb, ra, rb, alpha, beta, t - 1)
            - (q * r / alpha) * expansion(la - 1, lb, ra, rb, alpha, beta, t)
            + (t + 1) * expansion(la - 1, lb, ra, rb, alpha, beta, t + 1)
        )

    return (
        (1 / (2 * p)) * expansion(la, lb - 1, ra, rb, alpha, beta, t - 1)
        + (q * r / beta) * expansion(la, lb - 1, ra, rb, alpha, beta, t)
        + (t + 1) * expansion(la, lb - 1, ra, rb, alpha, beta, t + 1)
    )


def gaussian_overlap(la, lb, ra, rb, alpha, beta):
    r"""Compute overlap integral for two primitive Gaussian functions.

    The overlap integral between two Gaussian functions denoted by :math:`a` and :math:`b` can be
    computed as [`Helgaker (1995) p803 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]:

    .. math::

        S_{ab} = E^{ij} E^{kl} E^{mn} \left (\frac{\pi}{p}  \right )^{3/2},

    where :math:`E` is a coefficient that can be computed recursively, :math:`i-n` are the angular
    momentum quantum numbers corresponding to different Cartesian components and :math:`p` is
    computed from the exponents of the two Gaussian functions as :math:`p = \alpha + \beta`.

    Args:
        la (integer): angular momentum for the first Gaussian function
        lb (integer): angular momentum for the second Gaussian function
        ra (float): position vector of the first Gaussian function
        rb (float): position vector of the second Gaussian function
        alpha (array[float]): exponent of the first Gaussian function
        beta (array[float]): exponent of the second Gaussian function

    Returns:
        array[float]: overlap integral between primitive Gaussian functions

    **Example**

    >>> la, lb = (0, 0, 0), (0, 0, 0)
    >>> ra, rb = np.array([0., 0., 0.]), np.array([0., 0., 0.])
    >>> alpha = np.array([np.pi/2])
    >>> beta = np.array([np.pi/2])
    >>> o = gaussian_overlap(la, lb, ra, rb, alpha, beta)
    >>> o
    array([1.])
    """
    p = alpha + beta
    s = 1.0
    for i in range(3):
        s = s * anp.sqrt(anp.pi / p) * expansion(la[i], lb[i], ra[i], rb[i], alpha, beta, 0)
    return s


def overlap_integral(basis_a, basis_b, normalize=True):
    r"""Return a function that computes the overlap integral for two contracted Gaussian functions.

    Args:
        basis_a (~qchem.basis_set.BasisFunction): first basis function
        basis_b (~qchem.basis_set.BasisFunction): second basis function
        normalize (bool): if True, the basis functions get normalized

    Returns:
        function: function that computes the overlap integral

    **Example**

    >>> symbols  = ['H', 'H']
    >>> geometry = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]], requires_grad = False)
    >>> mol = qml.qchem.Molecule(symbols, geometry)
    >>> args = []
    >>> overlap_integral(mol.basis_set[0], mol.basis_set[0])(*args)
    1.0
    """

    def _overlap_integral(*args):
        r"""Normalize and compute the overlap integral for two contracted Gaussian functions.

        Args:
            args (array[float]): initial values of the differentiable parameters

        Returns:
            array[float]: the overlap integral between two contracted Gaussian orbitals
        """

        args_a = [arg[0] for arg in args]
        args_b = [arg[1] for arg in args]
        alpha, ca, ra = _generate_params(basis_a.params, args_a)
        beta, cb, rb = _generate_params(basis_b.params, args_b)

        if basis_a.params[1].requires_grad or normalize:
            ca = ca * primitive_norm(basis_a.l, alpha)
            cb = cb * primitive_norm(basis_b.l, beta)
            na = contracted_norm(basis_a.l, alpha, ca)
            nb = contracted_norm(basis_b.l, beta, cb)
        else:
            na = nb = 1.0

        return (
            na
            * nb
            * (
                (ca[:, anp.newaxis] * cb)
                * gaussian_overlap(basis_a.l, basis_b.l, ra, rb, alpha[:, anp.newaxis], beta)
            ).sum()
        )

    return _overlap_integral


def hermite_moment(alpha, beta, t, order, r):
    r"""Compute the Hermite moment integral recursively.

    The Hermite moment integral in one dimension is defined as

    .. math::

        M_{t}^{e} = \int_{-\infty }^{+\infty} q^e \Lambda_t dq,

    where :math:`e` is a positive integer, that is represented by the ``order`` argument,
    :math:`q = x, y, z` is the coordinate at which the integral is evaluatedand and
    :math:`\Lambda_t` is the :math:`t` component of the Hermite Gaussian function. The integral can
    be computed recursively as
    [`Helgaker (1995) p802 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]

    .. math::

        M_{t}^{e+1} = t M_{t-1}^{e} + Q M_{t}^{e} + \frac{1}{2p} M_{t+1}^{e},

    where :math:`Q` is the distance between the center of the Hermite Gaussian function and the
    origin, at dimension :math:`q = x, y, z` of the Cartesian coordinates system.

    This integral is zero for :math:`t > e` and the base case solution is

    .. math::

        M_t^0 = \delta _{t0} \sqrt{\frac{\pi}{p}},

    where :math:`p = \alpha + \beta` and :math:`\alpha, \beta` are the exponents of the Gaussian
    functions that construct the Hermite Gaussian function :math:`\Lambda`.

    Args:
        alpha (array[float]): exponent of the left Gaussian function
        beta (array[float]): exponent of the right Gaussian function
        t (integer): order of the Hermite Gaussian function
        order (integer): exponent of the position component
        r (array[float]): distance between the center of the Hermite Gaussian function and the origin

    Returns:
        array[float]: the Hermite moment integral

    **Example**

    >>> alpha = np.array([3.42525091])
    >>> beta = np.array([3.42525091])
    >>> t = 0
    >>> order = 1
    >>> r = 1.5
    >>> hermite_moment(alpha, beta, t, order, r)
    array([1.0157925])
    """
    p = anp.array(alpha + beta)

    if t > order or (order == 0 and t != 0):
        return 0.0
    if order == 0 and t == 0:
        return anp.sqrt(anp.pi / p)
    m = (
        hermite_moment(alpha, beta, t - 1, order - 1, r) * t
        + hermite_moment(alpha, beta, t, order - 1, r) * r
        + hermite_moment(alpha, beta, t + 1, order - 1, r) / (2 * p)
    )
    return m


def gaussian_moment(li, lj, ri, rj, alpha, beta, order, r):
    r"""Compute the one-dimensional multipole moment integral for two primitive Gaussian functions.

    The multipole moment integral in one dimension is defined as

    .. math::

        S_{ij}^e = \left \langle G_i | q^e | G_j \right \rangle,

    where :math:`G` is a Gaussian function at dimension :math:`q = x, y, z` of the Cartesian
    coordinates system and :math:`e` is a positive integer that is represented by the ``order``
    argument. The integrals can be evaluated as
    [`Helgaker (1995) p803 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]

    .. math::

        S_{ij}^e = \sum_{t=0}^{\mathrm{min}(i+j, \ e)} E_t^{ij} M_t^e,

    where :math:`E` and :math:`M` are the Hermite Gaussian expansion coefficient and the Hermite
    moment integral, respectively, that can be computed recursively.

    Args:
        li (integer): angular momentum for the left Gaussian function
        lj (integer): angular momentum for the right Gaussian function
        ri (float): position of the left Gaussian function
        rj (float): position of the right Gaussian function
        alpha (array[float]): exponent of the left Gaussian function
        beta (array[float]): exponent of the right Gaussian function
        order (integer): exponent of the position component
        r (array[float]): distance between the center of the Hermite Gaussian function and origin

    Returns:
        array[float]: one-dimensional multipole moment integral between primitive Gaussian functions

    **Example**

    >>> li, lj = 0, 0
    >>> ri, rj = np.array([2.0]), np.array([2.0])
    >>> alpha = np.array([3.42525091])
    >>> beta = np.array([3.42525091])
    >>> order = 1
    >>> r = 1.5
    >>> gaussian_moment(li, lj, ri, rj, alpha, beta, order, r)
    array([1.0157925])
    """
    s = 0.0
    for t in range(min(li + lj, order) + 1):
        s = s + expansion(li, lj, ri, rj, alpha, beta, t) * hermite_moment(alpha, beta, t, order, r)

    return s


def moment_integral(basis_a, basis_b, order, idx, normalize=True):
    r"""Return a function that computes the multipole moment integral for two contracted Gaussians.

    The multipole moment integral for two primitive Gaussian functions is computed as

    .. math::

        S^e = \left \langle G_i | q^e | G_j \right \rangle
                   \left \langle G_k | G_l \right \rangle
                   \left \langle G_m | G_n \right \rangle,

    where :math:`G_{i-n}` is a one-dimensional Gaussian function, :math:`q = x, y, z` is the
    coordinate at which the integral is evaluated and :math:`e` is a positive integer that is
    represented by the ``order`` argument. For contracted Gaussians, these integrals will be
    computed over primitive Gaussians, multiplied by the normalized contraction coefficients and
    finally summed over.

    The ``idx`` argument determines the coordinate :math:`q` at which the integral is computed. It
    can be :math:`0, 1, 2` for :math:`x, y, z` components, respectively.

    Args:
        basis_a (~qchem.basis_set.BasisFunction): left basis function
        basis_b (~qchem.basis_set.BasisFunction): right basis function
        order (integer): exponent of the position component
        idx (integer): index determining the dimension of the multipole moment integral
        normalize (bool): if True, the basis functions get normalized

    Returns:
        function: function that computes the multipole moment integral

    **Example**

    >>> symbols  = ['H', 'Li']
    >>> geometry = np.array([[0.0, 0.0, 0.0], [2.0, 0.0, 0.0]], requires_grad = True)
    >>> mol = qml.qchem.Molecule(symbols, geometry)
    >>> args = [mol.r] # initial values of the differentiable parameters
    >>> order, idx =  1, 0
    >>> moment_integral(mol.basis_set[0], mol.basis_set[1], order, idx)(*args)
    3.12846324e-01
    """

    def _moment_integral(*args):
        r"""Normalize and compute the multipole moment integral for two contracted Gaussians.

        Args:
            args (array[float]): initial values of the differentiable parameters

        Returns:
            array[float]: the multipole moment integral between two contracted Gaussian orbitals
        """
        args_a = [arg[0] for arg in args]
        args_b = [arg[1] for arg in args]

        la = basis_a.l
        lb = basis_b.l

        alpha, ca, ra = _generate_params(basis_a.params, args_a)
        beta, cb, rb = _generate_params(basis_b.params, args_b)

        if basis_a.params[1].requires_grad or normalize:
            ca = ca * primitive_norm(basis_a.l, alpha)
            cb = cb * primitive_norm(basis_b.l, beta)
            na = contracted_norm(basis_a.l, alpha, ca)
            nb = contracted_norm(basis_b.l, beta, cb)
        else:
            na = nb = 1.0

        p = alpha[:, anp.newaxis] + beta
        q = anp.sqrt(anp.pi / p)
        r = (
            alpha[:, anp.newaxis] * ra[:, anp.newaxis, anp.newaxis]
            + beta * rb[:, anp.newaxis, anp.newaxis]
        ) / p

        i, j, k = anp.roll(anp.array([0, 2, 1]), idx)

        s = (
            gaussian_moment(la[i], lb[i], ra[i], rb[i], alpha[:, anp.newaxis], beta, order, r[i])
            * expansion(la[j], lb[j], ra[j], rb[j], alpha[:, anp.newaxis], beta, 0)
            * q
            * expansion(la[k], lb[k], ra[k], rb[k], alpha[:, anp.newaxis], beta, 0)
            * q
        )

        return (na * nb * (ca[:, anp.newaxis] * cb) * s).sum()

    return _moment_integral


def _diff2(i, j, ri, rj, alpha, beta):
    r"""Compute the second order differentiated integral needed for evaluating a kinetic integral.

    The second-order integral :math:`D_{ij}^2`, where :math:`i` and :math:`j` denote angular
    momentum components of Gaussian functions, is computed from overlap integrals :math:`S` and the
    Gaussian exponent :math:`\beta` as
    [`Helgaker (1995) p804 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]:

    .. math::

        D_{ij}^2 = j(j-1)S_{i,j-2}^0 - 2\beta(2j+1)S_{i,j}^0 + 4\beta^2 S_{i,j+2}^0.

    Args:
        i (integer): angular momentum component for the first Gaussian function
        j (integer): angular momentum component for the second Gaussian function
        ri (array[float]): position component of the the first Gaussian function
        rj (array[float]): position component of the the second Gaussian function
        alpha (array[float]): exponent of the first Gaussian function
        beta (array[float]): exponent of the second Gaussian function

    Returns:
        array[float]: second-order differentiated integral between two Gaussian functions
    """
    p = alpha + beta

    d1 = j * (j - 1) * anp.sqrt(anp.pi / p) * expansion(i, j - 2, ri, rj, alpha, beta, 0)
    d2 = -2 * beta * (2 * j + 1) * anp.sqrt(anp.pi / p) * expansion(i, j, ri, rj, alpha, beta, 0)
    d3 = 4 * beta**2 * anp.sqrt(anp.pi / p) * expansion(i, j + 2, ri, rj, alpha, beta, 0)

    return d1 + d2 + d3


def gaussian_kinetic(la, lb, ra, rb, alpha, beta):
    r"""Compute the kinetic integral for two primitive Gaussian functions.

    The kinetic integral between two Gaussian functions denoted by :math:`a` and :math:`b` is
    computed as
    [`Helgaker (1995) p805 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]:

    .. math::

        T_{ab} = -\frac{1}{2} \left ( D_{ij}^2 D_{kl}^0 D_{mn}^0 + D_{ij}^0 D_{kl}^2 D_{mn}^0 + D_{ij}^0 D_{kl}^0 D_{mn}^2\right ),

    where :math:`D_{ij}^0 = S_{ij}^0` is an overlap integral and :math:`D_{ij}^2` is computed from
    overlap integrals :math:`S` and the Gaussian exponent :math:`\beta` as

    .. math::

        D_{ij}^2 = j(j-1)S_{i,j-2}^0 - 2\beta(2j+1)S_{i,j}^0 + 4\beta^2 S_{i,j+2}^0.

    Args:
        la (tuple[int]): angular momentum for the first Gaussian function
        lb (tuple[int]): angular momentum for the second Gaussian function
        ra (array[float]): position vector of the the first Gaussian function
        rb (array[float]): position vector of the the second Gaussian function
        alpha (array[float]): exponent of the first Gaussian function
        beta (array[float]): exponent of the second Gaussian function

    Returns:
        array[float]: kinetic integral between two Gaussian functions

    **Example**

    >>> la, lb = (0, 0, 0), (0, 0, 0)
    >>> ra = np.array([0., 0., 0.])
    >>> rb = rb = np.array([0., 0., 0.])
    >>> alpha = np.array([np.pi/2])
    >>> beta = np.array([np.pi/2])
    >>> t = gaussian_kinetic(la, lb, ra, rb, alpha, beta)
    >>> t
    array([2.35619449])
    """

    p = alpha + beta

    t1 = (
        _diff2(la[0], lb[0], ra[0], rb[0], alpha, beta)
        * anp.sqrt(anp.pi / p)
        * expansion(la[1], lb[1], ra[1], rb[1], alpha, beta, 0)
        * anp.sqrt(anp.pi / p)
        * expansion(la[2], lb[2], ra[2], rb[2], alpha, beta, 0)
    )

    t2 = (
        anp.sqrt(anp.pi / p)
        * expansion(la[0], lb[0], ra[0], rb[0], alpha, beta, 0)
        * _diff2(la[1], lb[1], ra[1], rb[1], alpha, beta)
        * anp.sqrt(anp.pi / p)
        * expansion(la[2], lb[2], ra[2], rb[2], alpha, beta, 0)
    )

    t3 = (
        anp.sqrt(anp.pi / p)
        * expansion(la[0], lb[0], ra[0], rb[0], alpha, beta, 0)
        * anp.sqrt(anp.pi / p)
        * expansion(la[1], lb[1], ra[1], rb[1], alpha, beta, 0)
        * _diff2(la[2], lb[2], ra[2], rb[2], alpha, beta)
    )

    return -0.5 * (t1 + t2 + t3)


def kinetic_integral(basis_a, basis_b, normalize=True):
    r"""Return a function that computes the kinetic integral for two contracted Gaussian functions.

    Args:
        basis_a (~qchem.basis_set.BasisFunction): first basis function
        basis_b (~qchem.basis_set.BasisFunction): second basis function
        normalize (bool): if True, the basis functions get normalized

    Returns:
        function: function that computes the kinetic integral

    **Example**

    >>> symbols  = ['H', 'H']
    >>> geometry = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]], requires_grad = False)
    >>> alpha = np.array([[3.425250914, 0.6239137298, 0.168855404],
    >>>                   [3.425250914, 0.6239137298, 0.168855404]], requires_grad = True)
    >>> mol = qml.qchem.Molecule(symbols, geometry, alpha=alpha)
    >>> args = [mol.alpha]
    >>> kinetic_integral(mol.basis_set[0], mol.basis_set[1])(*args)
    0.38325367405312843
    """

    def _kinetic_integral(*args):
        r"""Compute the kinetic integral for two contracted Gaussian functions.

        Args:
            args (array[float]): initial values of the differentiable parameters

        Returns:
            array[float]: the kinetic integral between two contracted Gaussian orbitals
        """
        args_a = [arg[0] for arg in args]
        args_b = [arg[1] for arg in args]
        alpha, ca, ra = _generate_params(basis_a.params, args_a)
        beta, cb, rb = _generate_params(basis_b.params, args_b)

        if basis_a.params[1].requires_grad or normalize:
            ca = ca * primitive_norm(basis_a.l, alpha)
            cb = cb * primitive_norm(basis_b.l, beta)
            na = contracted_norm(basis_a.l, alpha, ca)
            nb = contracted_norm(basis_b.l, beta, cb)
        else:
            na = nb = 1.0

        return (
            na
            * nb
            * (
                (ca[:, anp.newaxis] * cb)
                * gaussian_kinetic(basis_a.l, basis_b.l, ra, rb, alpha[:, anp.newaxis], beta)
            ).sum()
        )

    return _kinetic_integral


def _boys(n, t):
    r"""Evaluate the Boys function.

    The :math:`n`-th order `Boys function <https://arxiv.org/abs/2107.01488>`_ is defined as

    .. math::

        F_n(t) = \int_{0}^{1}x^{2n} e^{-tx^2}dx.

    The Boys function is related to the lower incomplete Gamma
    `function <https://en.wikipedia.org/wiki/Incomplete_gamma_function>`_, :math:`\gamma`, as

    .. math::

        F_n(t) = \frac{1}{2t^{n + 0.5}} \gamma(n + 0.5, t),

    where

    .. math::

        \gamma(m, t) = \int_{0}^{t} x^{m-1} e^{-x} dx.

    Args:
        n (float): order of the Boys function
        t (array[float]): exponent of the Boys function

    Returns:
        (array[float]): value of the Boys function
    """
    return anp.where(
        t == 0.0,
        1 / (2 * n + 1),
        asp.special.gammainc(n + 0.5, t + (t == 0.0))
        * asp.special.gamma(n + 0.5)
        / (2 * (t + (t == 0.0)) ** (n + 0.5)),
    )  # (t == 0.0) is added to avoid divide by zero


def _hermite_coulomb(t, u, v, n, p, dr):
    """Evaluate the Hermite integral needed to compute the nuclear attraction and electron repulsion
    integrals.

    These integrals are computed recursively starting from the Boys function
    [`Helgaker (1995) p817 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]:


    .. math::

        R_{000}^n = (-2p)^n F_n(pR_{CP}^2),

    where :math:`F_n` is the Boys function, :math:`p` is computed from the exponents of the two
    Gaussian functions as :math:`p = \alpha + \beta`, and :math:`R_{CP}` is the distance between the
    center of the composite Gaussian centered at :math:`P` and the electrostatic potential generated
    by a nucleus at :math:`C`. The following recursive equations are used to compute the
    higher-order Hermite integrals

    .. math::

        R_{t+1, u, v}^n = t R_{t-1, u, v}^{n+1} + x R_{t, u, v}^{n+1},

        R_{t, u+1, v}^n = u R_{t, u-1, v}^{n+1} + y R_{t, u, v}^{n+1},

        R_{t, u, v+1}^n = v R_{t, u, v-1}^{n+1} + z R_{t, u, v}^{n+1},

    where :math:`x`, :math:`y` and :math:`z` are the Cartesian components of :math:`R_{CP}`.

    Args:
        t (integer): order of Hermite derivative in x
        u (integer): order of Hermite derivative in y
        v (float): order of Hermite derivative in z
        n (integer): order of the Boys function
        p (float): sum of the Gaussian exponents
        dr (array[float]): distance between the center of the composite Gaussian and the nucleus

    Returns:
        array[float]: value of the Hermite integral
    """
    x, y, z = dr[0], dr[1], dr[2]
    T = p * (dr**2).sum(axis=0)
    r = 0

    if t == u == v == 0:
        return ((-2 * p) ** n) * _boys(n, T)

    if t == u == 0:
        if v > 1:
            r = r + (v - 1) * _hermite_coulomb(t, u, v - 2, n + 1, p, dr)
        r = r + z * _hermite_coulomb(t, u, v - 1, n + 1, p, dr)
        return r

    if t == 0:
        if u > 1:
            r = r + (u - 1) * _hermite_coulomb(t, u - 2, v, n + 1, p, dr)
        r = r + y * _hermite_coulomb(t, u - 1, v, n + 1, p, dr)
        return r

    if t > 1:
        r = r + (t - 1) * _hermite_coulomb(t - 2, u, v, n + 1, p, dr)
    r = r + x * _hermite_coulomb(t - 1, u, v, n + 1, p, dr)
    return r


def nuclear_attraction(la, lb, ra, rb, alpha, beta, r):
    r"""Compute nuclear attraction integral between primitive Gaussian functions.

    The nuclear attraction integral between two Gaussian functions denoted by :math:`a` and
    :math:`b` can be computed as
    [`Helgaker (1995) p820 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]

    .. math::

        V_{ab} = \frac{2\pi}{p} \sum_{tuv} E_t^{ij} E_u^{kl} E_v^{mn} R_{tuv},

    where :math:`E` and :math:`R` represent the Hermite Gaussian expansion coefficients and the
    Hermite Coulomb integral, respectively. The sum goes over :math:`i + j + 1`, :math:`k + l + 1`
    and :math:`m + n + 1` for :math:`t`, :math:`u` and :math:`v`, respectively, and :math:`p` is
    computed from the exponents of the two Gaussian functions as :math:`p = \alpha + \beta`.

    Args:
        la (tuple[int]): angular momentum for the first Gaussian function
        lb (tuple[int]): angular momentum for the second Gaussian function
        ra (array[float]): position vector of the the first Gaussian function
        rb (array[float]): position vector of the the second Gaussian function
        alpha (array[float]): exponent of the first Gaussian function
        beta (array[float]): exponent of the second Gaussian function
        r (array[float]): position vector of nucleus

    Returns:
        array[float]: nuclear attraction integral between two Gaussian functions
    """
    l1, m1, n1 = la
    l2, m2, n2 = lb
    p = alpha + beta
    rgp = (alpha * ra[:, anp.newaxis, anp.newaxis] + beta * rb[:, anp.newaxis, anp.newaxis]) / (
        alpha + beta
    )
    dr = rgp - anp.array(r)[:, anp.newaxis, anp.newaxis]

    a = 0.0
    for t, u, v in it.product(*[range(l) for l in [l1 + l2 + 1, m1 + m2 + 1, n1 + n2 + 1]]):
        a = a + expansion(l1, l2, ra[0], rb[0], alpha, beta, t) * expansion(
            m1, m2, ra[1], rb[1], alpha, beta, u
        ) * expansion(n1, n2, ra[2], rb[2], alpha, beta, v) * _hermite_coulomb(t, u, v, 0, p, dr)
    a = a * 2 * anp.pi / p
    return a


def attraction_integral(r, basis_a, basis_b, normalize=True):
    r"""Return a function that computes the nuclear attraction integral for two contracted Gaussian
    functions.

    Args:
        r (array[float]): position vector of nucleus
        basis_a (~qchem.basis_set.BasisFunction): first basis function
        basis_b (~qchem.basis_set.BasisFunction): second basis function
        normalize (bool): if True, the basis functions get normalized

    Returns:
        function: function that computes the electron-nuclear attraction integral

    **Example**

    >>> symbols  = ['H', 'H']
    >>> geometry = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]], requires_grad = False)
    >>> alpha = np.array([[3.425250914, 0.6239137298, 0.168855404],
    >>>                   [3.425250914, 0.6239137298, 0.168855404]], requires_grad = True)
    >>> mol = qml.qchem.Molecule(symbols, geometry, alpha=alpha)
    >>> basis_a = mol.basis_set[0]
    >>> basis_b = mol.basis_set[1]
    >>> args = [mol.alpha]
    >>> attraction_integral(geometry[0], basis_a, basis_b)(*args)
    0.801208332328965
    """

    def _attraction_integral(*args):
        r"""Compute the electron-nuclear attraction integral for two contracted Gaussian functions.

        Args:
            args (array[float]): initial values of the differentiable parameters

        Returns:
            array[float]: the electron-nuclear attraction integral
        """
        if r.requires_grad:
            coor = args[0]
            args_a = [arg[0] for arg in args[1:]]
            args_b = [arg[1] for arg in args[1:]]
        else:
            coor = r
            args_a = [arg[0] for arg in args]
            args_b = [arg[1] for arg in args]

        alpha, ca, ra = _generate_params(basis_a.params, args_a)
        beta, cb, rb = _generate_params(basis_b.params, args_b)

        if basis_a.params[1].requires_grad or normalize:
            ca = ca * primitive_norm(basis_a.l, alpha)
            cb = cb * primitive_norm(basis_b.l, beta)
            na = contracted_norm(basis_a.l, alpha, ca)
            nb = contracted_norm(basis_b.l, beta, cb)
        else:
            na = nb = 1.0

        v = (
            na
            * nb
            * (
                (ca * cb[:, anp.newaxis])
                * nuclear_attraction(
                    basis_a.l, basis_b.l, ra, rb, alpha, beta[:, anp.newaxis], coor
                )
            ).sum()
        )
        return v

    return _attraction_integral


def electron_repulsion(la, lb, lc, ld, ra, rb, rc, rd, alpha, beta, gamma, delta):
    r"""Compute the electron-electron repulsion integral between four primitive Gaussian functions.

    The electron repulsion integral between four Gaussian functions denoted by :math:`a`, :math:`b`
    , :math:`c` and :math:`d` is computed as
    [`Helgaker (1995) p820 <https://www.worldscientific.com/doi/abs/10.1142/9789812832115_0001>`_]

    .. math::

        g_{abcd} = \frac{2\pi^{5/2}}{pq\sqrt{p+q}} \sum_{tuv} E_t^{o_a o_b} E_u^{m_a m_b}
        E_v^{n_a n_b} \sum_{rsw} (-1)^{r+s+w} E_r^{o_c o_d} E_s^{m_c m_d} E_w^{n_c n_d}
        R_{t+r, u+s, v+w},

    where :math:`E` and :math:`R` are the Hermite Gaussian expansion coefficients and the
    Hermite Coulomb integral, respectively. The sums go over the angular momentum quantum numbers
    :math:`o_i + o_j + 1`, :math:`m_i + m_j + 1` and :math:`n_i + n_j + 1` respectively for
    :math:`t, u, v` and :math:`r, s, w`. The exponents of the Gaussian functions are used to compute
    :math:`p` and :math:`q` as :math:`p = \alpha + \beta` and :math:`q = \gamma + \delta`.

    Args:
        la (tuple[int]): angular momentum for the first Gaussian function
        lb (tuple[int]): angular momentum for the second Gaussian function
        lc (tuple[int]): angular momentum for the third Gaussian function
        ld (tuple[int]): angular momentum for the forth Gaussian function
        ra (array[float]): position vector of the the first Gaussian function
        rb (array[float]): position vector of the the second Gaussian function
        rc (array[float]): position vector of the the third Gaussian function
        rd (array[float]): position vector of the the forth Gaussian function
        alpha (array[float]): exponent of the first Gaussian function
        beta (array[float]): exponent of the second Gaussian function
        gamma (array[float]): exponent of the third Gaussian function
        delta (array[float]): exponent of the forth Gaussian function

    Returns:
        array[float]: electron-electron repulsion integral between four Gaussian functions
    """
    l1, m1, n1 = la
    l2, m2, n2 = lb
    l3, m3, n3 = lc
    l4, m4, n4 = ld

    p = alpha + beta
    q = gamma + delta

    p_ab = (
        alpha * ra[:, anp.newaxis, anp.newaxis, anp.newaxis, anp.newaxis]
        + beta * rb[:, anp.newaxis, anp.newaxis, anp.newaxis, anp.newaxis]
    ) / (alpha + beta)

    p_cd = (
        gamma * rc[:, anp.newaxis, anp.newaxis, anp.newaxis, anp.newaxis]
        + delta * rd[:, anp.newaxis, anp.newaxis, anp.newaxis, anp.newaxis]
    ) / (gamma + delta)

    g_t = [expansion(l1, l2, ra[0], rb[0], alpha, beta, t) for t in range(l1 + l2 + 1)]
    g_u = [expansion(m1, m2, ra[1], rb[1], alpha, beta, u) for u in range(m1 + m2 + 1)]
    g_v = [expansion(n1, n2, ra[2], rb[2], alpha, beta, v) for v in range(n1 + n2 + 1)]
    g_r = [expansion(l3, l4, rc[0], rd[0], gamma, delta, r) for r in range(l3 + l4 + 1)]
    g_s = [expansion(m3, m4, rc[1], rd[1], gamma, delta, s) for s in range(m3 + m4 + 1)]
    g_w = [expansion(n3, n4, rc[2], rd[2], gamma, delta, w) for w in range(n3 + n4 + 1)]

    g = 0.0
    lengths = [l1 + l2 + 1, m1 + m2 + 1, n1 + n2 + 1, l3 + l4 + 1, m3 + m4 + 1, n3 + n4 + 1]
    for t, u, v, r, s, w in it.product(*[range(length) for length in lengths]):
        g = g + g_t[t] * g_u[u] * g_v[v] * g_r[r] * g_s[s] * g_w[w] * (
            (-1) ** (r + s + w)
        ) * _hermite_coulomb(t + r, u + s, v + w, 0, (p * q) / (p + q), p_ab - p_cd)

    g = g * 2 * (anp.pi**2.5) / (p * q * anp.sqrt(p + q))

    return g


def repulsion_integral(basis_a, basis_b, basis_c, basis_d, normalize=True):
    r"""Return a function that computes the electron-electron repulsion integral for four contracted
    Gaussian functions.

    Args:
        basis_a (~qchem.basis_set.BasisFunction): first basis function
        basis_b (~qchem.basis_set.BasisFunction): second basis function
        basis_c (~qchem.basis_set.BasisFunction): third basis function
        basis_d (~qchem.basis_set.BasisFunction): fourth basis function
        normalize (bool): if True, the basis functions get normalized

    Returns:
        function: function that computes the electron repulsion integral

    **Example**

    >>> symbols  = ['H', 'H']
    >>> geometry = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]], requires_grad = False)
    >>> alpha = np.array([[3.425250914, 0.6239137298, 0.168855404],
    >>>                   [3.425250914, 0.6239137298, 0.168855404],
    >>>                   [3.425250914, 0.6239137298, 0.168855404],
    >>>                   [3.425250914, 0.6239137298, 0.168855404]], requires_grad = True)
    >>> mol = qml.qchem.Molecule(symbols, geometry, alpha=alpha)
    >>> basis_a = mol.basis_set[0]
    >>> basis_b = mol.basis_set[1]
    >>> args = [mol.alpha]
    >>> repulsion_integral(basis_a, basis_b, basis_a, basis_b)(*args)
    0.45590152106593573
    """

    def _repulsion_integral(*args):
        r"""Compute the electron-electron repulsion integral for four contracted Gaussian functions.

        Args:
            args (array[float]): initial values of the differentiable parameters

        Returns:
            array[float]: the electron repulsion integral between four contracted Gaussian functions
        """
        args_a = [arg[0] for arg in args]
        args_b = [arg[1] for arg in args]
        args_c = [arg[2] for arg in args]
        args_d = [arg[3] for arg in args]

        alpha, ca, ra = _generate_params(basis_a.params, args_a)
        beta, cb, rb = _generate_params(basis_b.params, args_b)
        gamma, cc, rc = _generate_params(basis_c.params, args_c)
        delta, cd, rd = _generate_params(basis_d.params, args_d)

        if basis_a.params[1].requires_grad or normalize:
            ca = ca * primitive_norm(basis_a.l, alpha)
            cb = cb * primitive_norm(basis_b.l, beta)
            cc = cc * primitive_norm(basis_c.l, gamma)
            cd = cd * primitive_norm(basis_d.l, delta)

            n1 = contracted_norm(basis_a.l, alpha, ca)
            n2 = contracted_norm(basis_b.l, beta, cb)
            n3 = contracted_norm(basis_c.l, gamma, cc)
            n4 = contracted_norm(basis_d.l, delta, cd)
        else:
            n1 = n2 = n3 = n4 = 1.0

        e = (
            n1
            * n2
            * n3
            * n4
            * (
                (
                    ca
                    * cb[:, anp.newaxis]
                    * cc[:, anp.newaxis, anp.newaxis]
                    * cd[:, anp.newaxis, anp.newaxis, anp.newaxis]
                )
                * electron_repulsion(
                    basis_a.l,
                    basis_b.l,
                    basis_c.l,
                    basis_d.l,
                    ra,
                    rb,
                    rc,
                    rd,
                    alpha,
                    beta[:, anp.newaxis],
                    gamma[:, anp.newaxis, anp.newaxis],
                    delta[:, anp.newaxis, anp.newaxis, anp.newaxis],
                )
            ).sum()
        )
        return e

    return _repulsion_integral



# Copyright 2018-2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the :class:`~pennylane.data.Dataset` class and its associated functions.
"""

from abc import ABC
from glob import glob
import os

from pennylane import Hamiltonian
from pennylane.pauli import string_to_pauli_word


def _import_zstd_dill():
    """Import zstd and dill."""
    try:
        # pylint: disable=import-outside-toplevel, unused-import, multiple-imports
        import zstd, dill
    except ImportError as Error:
        raise ImportError(
            "This feature requires zstd and dill."
            "They can be installed with:\n\n pip install zstd dill."
        ) from Error

    return zstd, dill


# pylint: disable=too-many-instance-attributes
class Dataset(ABC):
    """Create a dataset object to store a collection of information describing
    a physical system and its evolution. For example, a dataset for an arbitrary
    quantum system could have a Hamiltonian, its ground state,
    and an efficient state-preparation circuit for that state.

    Args:
        *args: For internal use only. These will be ignored if called with ``standard=False``
        standard (bool): For internal use only. See the note below for the behavior when this is set to ``True``
        **kwargs: variable-length keyword arguments specifying the data to be stored in the dataset

    Note on the ``standard`` kwarg:
        A `standard` Dataset uses previously generated, hosted quantum data. This special instance of the
        ``Dataset`` class makes certain assumptions about folder management for downloading the data and
        handling I/O. As such, the ``Dataset`` class should not be instantiated by the users directly with
        ``standard=True``. Instead, they should use :meth:`~load`.


    .. seealso:: :meth:`~load`

    **Example**

    >>> Hamiltonian = qml.Hamiltonian([1., 1.], [qml.PauliZ(wires=0), qml.PauliZ(wires=1)])
    >>> eigvals, eigvecs = np.linalg.eigh(qml.matrix(Hamiltonian))
    >>> ground_state_energy = np.min(eigvals)
    >>> ground_state = np.transpose(eigvecs)[np.argmin(eigvals)]
    >>> dataset = qml.data.Dataset(Hamiltonian = Hamiltonian, ground_state = ground_state,
            ground_state_energy = ground_state_energy)
    >>> print(dataset.Hamiltonian)
          (1) [Z0]
        + (1) [Z1]
    >>> print(dataset.ground_energy)
    -2.0

    .. details::
        :title: Usage Details

        In addition to creating datasets in memory, we can also store them in
        the disk and then load them as follows. First we create the dataset:

        >>> Hamiltonian = qml.Hamiltonian([1., 1.], [qml.PauliZ(wires=0), qml.PauliZ(wires=1)])
        ...     eigvals, eigvecs = np.linalg.eigh(qml.matrix(Hamiltonian))
        >>> ground_state_energy = np.min(eigvals)
        >>> ground_state = np.transpose(eigvecs)[np.argmin(eigvals)]
        >>> dataset = qml.data.Dataset(Hamiltonian = Hamiltonian, ground_state = ground_state,
        ...     ground_state_energy = ground_state_energy)

        Then to save the dataset to a file, we call :func:`Dataset.write()`:

        >>> dataset.write('./path/to/file/dataset.dat')

        We can then retrieve the data using :func:`Dataset.read()`

        >>> retrieved_data = qml.data.Dataset()
        >>> retrieved_data.read('./path/to/file/dataset.dat')
        >>> print(retrieved_data.Hamiltonian)
          (1) [Z0]
        + (1) [Z1]
        >>> print(dataset.Hamiltonian)
          (1) [Z0]
        + (1) [Z1]
    """

    _standard_argnames = ["data_name", "data_folder", "attr_prefix", "docstring"]

    def __std_init__(self, data_name, folder, attr_prefix, docstring):
        """Constructor for standardized datasets."""
        self._dtype = data_name
        self._folder = folder.rstrip(os.path.sep)
        self._prefix = os.path.join(self._folder, attr_prefix) + "_{}.dat"
        self._prefix_len = len(attr_prefix) + 1
        self._description = os.path.join(data_name, self._folder.split(data_name)[-1][1:])
        self.__doc__ = docstring

        self._fullfile = self._prefix.format("full")
        if not os.path.exists(self._fullfile):
            self._fullfile = None

        for f in glob(f"{self._folder}{os.path.sep}*.dat"):
            self.read(f, lazy=True)

    def __base_init__(self, **kwargs):
        """Constructor for user-defined datasets."""
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __init__(self, *args, standard=False, **kwargs):
        """Dispatching constructor for direct invocation with kwargs or from qml.data.load()"""
        if standard:
            if len(args) != len(self._standard_argnames):
                raise TypeError(
                    f"Standard datasets expect {len(self._standard_argnames)} arguments: {self._standard_argnames}"
                )
            for name, val in zip(self._standard_argnames, args):
                if not isinstance(val, str):
                    raise ValueError(f"Expected {name} to be a str, got {type(val).__name__}")
            self._is_standard = True
            self.__std_init__(*args)
        else:
            self._is_standard = False
            self.__base_init__(**kwargs)

    def __repr__(self):
        attr_str = (
            str(list(self.attrs))
            if len(self.attrs) < 3
            else f"{str(list(self.attrs)[:2])[:-1]}, ...]"
        )

        std_str = f"description: {self._description}, " if self._is_standard else ""
        return f"<Dataset = {std_str}attributes: {attr_str}>"

    @property
    def attrs(self):
        """Returns attributes of the dataset."""
        return {k: v for k, v in vars(self).items() if k[0] != "_"}

    @staticmethod
    def _read_file(filepath):
        """Read data from a saved file.

        Returns:
            A data value for non-standard datasets or full files, otherwise a dictionary
        """
        with open(filepath, "rb") as f:
            compressed_pickle = f.read()

        zstd, dill = _import_zstd_dill()
        depressed_pickle = zstd.decompress(compressed_pickle)
        return dill.loads(depressed_pickle)

    def read(self, filepath, lazy=False):
        """Loads data from a saved file to the current dataset.

        Args:
            filepath (string): The desired location and filename to load, e.g. './path/to/file/file_name.dat'.
            lazy (bool): Indicates if only the key of the attribute should be saved to the Dataset instance

        **Example**

        >>> new_dataset = qml.data.Dataset(kw1 = 1, kw2 = '2', kw3 = [3])
        >>> new_dataset.read('./path/to/file/file_name.dat')
        """
        # set 'full' for non-standard datasets so they read keys too
        attribute = self.__get_attribute_from_filename(filepath) if self._is_standard else "full"
        if attribute == "full":
            data = self._read_file(filepath)
            for attr, value in data.items():
                setattr(self, f"{attr}", None if lazy else value)
        else:
            data = None if lazy else self._read_file(filepath)
            setattr(self, f"{attribute}", data)

    @staticmethod
    def _write_file(data, filepath, protocol=4):
        """General method to write data to a file."""
        zstd, dill = _import_zstd_dill()
        pickled_data = dill.dumps(data, protocol=protocol)
        compressed_pickle = zstd.compress(pickled_data)
        with open(filepath, "wb") as f:
            f.write(compressed_pickle)

    def write(self, filepath, protocol=4):
        """Writes the dataset to a file as a dictionary.

        Args:
            filepath (string): the desired save location and file name

        **Example**

        >>> new_dataset = qml.data.Dataset(kw1 = 1, kw2 = '2', kw3 = [3])
        >>> new_dataset.write('./path/to/file/file_name.dat')
        """
        dirname = os.path.dirname(filepath)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        self._write_file(self.attrs, filepath, protocol=protocol)

    def list_attributes(self):
        """List the attributes saved on the Dataset"""
        return list(self.attrs)

    # pylint: disable=attribute-defined-outside-init
    def __copy__(self):
        cls = self.__class__

        if not self._is_standard:
            return cls(**self.attrs)

        copied = cls.__new__(cls)
        copied._is_standard = True
        copied._dtype = self._dtype
        copied._folder = self._folder
        copied._prefix = self._prefix
        copied._prefix_len = self._prefix_len
        copied._fullfile = self._fullfile
        copied.__doc__ = self.__doc__
        for key, val in self.attrs.items():
            setattr(copied, f"{key}", val)

        return copied

    # The methods below are intended for use only by standard Dataset objects
    def __get_filename_for_attribute(self, attribute):
        return self._fullfile or self._prefix.format(attribute)

    def __get_attribute_from_filename(self, filename):
        return os.path.basename(filename)[self._prefix_len : -4]

    @staticmethod
    def __dict_to_hamiltonian(hamil_dict, wire_map):
        coeffs, ops = [], []
        for key, val in hamil_dict.items():
            coeffs.append(val)
            ops.append(string_to_pauli_word(key, wire_map))
        return Hamiltonian(coeffs, ops)

    # TODO: validate this and add to _write_file (if 'hamiltonian' in data))
    # @staticmethod
    # def __hamiltonian_to_dict(hamiltonian, wire_map):
    #     hamil_dict = {}
    #     coeffs, ops = hamiltonian.terms()
    #     for coeff, op in zip(coeffs, ops):
    #         hamil_dict.update({pauli_word_to_string(op, wire_map): coeff})
    #     return hamil_dict

    def __getattribute__(self, name):
        try:
            val = super().__getattribute__(name)
            if val is None and self._is_standard and name in self.attrs:
                raise AttributeError(
                    f"Dataset has a '{name}' attribute, but it is None and no data file was found"
                )
            return val
        except AttributeError:
            if not self._is_standard:
                raise
            filepath = self.__get_filename_for_attribute(name)
            if not os.path.exists(filepath):  # TODO: download the file here?
                raise
            value = self._read_file(filepath)
            if filepath == self._fullfile:
                if name not in value:
                    raise
                value = value[name]
            if name == "hamiltonian":
                value = self.__dict_to_hamiltonian(value["terms"], value["wire_map"])
            setattr(self, name, value)
            return value



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for circuit optimization."""
# pylint: disable=too-many-return-statements,import-outside-toplevel
from functools import partial

import numpy as np

from pennylane.math import abs as math_abs
from pennylane.math import allclose, arccos, arctan2, cos, get_interface, is_abstract, sin, stack
from pennylane.wires import Wires


def find_next_gate(wires, op_list):
    """Given a list of operations, finds the next operation that acts on at least one of
    the same set of wires, if present.

    Args:
        wires (Wires): A set of wires acted on by a quantum operation.
        op_list (list[Operation]): A list of operations that are implemented after the
            operation that acts on ``wires``.

    Returns:
        int or None: The index, in ``op_list``, of the earliest gate that uses one or more
        of the same wires, or ``None`` if no such gate is present.
    """
    next_gate_idx = None

    for op_idx, op in enumerate(op_list):
        if len(Wires.shared_wires([wires, op.wires])) > 0:
            next_gate_idx = op_idx
            break

    return next_gate_idx


def _zyz_to_quat(angles):
    """Converts a set of Euler angles in ZYZ format to a quaternion."""
    qw = cos(angles[1] / 2) * cos(0.5 * (angles[0] + angles[2]))
    qx = -sin(angles[1] / 2) * sin(0.5 * (angles[0] - angles[2]))
    qy = sin(angles[1] / 2) * cos(0.5 * (angles[0] - angles[2]))
    qz = cos(angles[1] / 2) * sin(0.5 * (angles[0] + angles[2]))

    return stack([qw, qx, qy, qz])


def _quaternion_product(q1, q2):
    """Compute the product of two quaternions, q = q1 * q2."""
    qw = q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2] - q1[3] * q2[3]
    qx = q1[0] * q2[1] + q1[1] * q2[0] + q1[2] * q2[3] - q1[3] * q2[2]
    qy = q1[0] * q2[2] - q1[1] * q2[3] + q1[2] * q2[0] + q1[3] * q2[1]
    qz = q1[0] * q2[3] + q1[1] * q2[2] - q1[2] * q2[1] + q1[3] * q2[0]

    return stack([qw, qx, qy, qz])


def _singular_quat_to_zyz(qw, qx, qy, qz, y_arg, abstract_jax=False):
    """Compute the ZYZ angles for the singular case of qx = qy = 0"""
    # pylint: disable=too-many-arguments
    z1_arg1 = 2 * (qx * qy + qz * qw)
    z1_arg2 = 1 - 2 * (qx**2 + qz**2)

    if abstract_jax:
        from jax.lax import cond

        return cond(
            y_arg > 0,
            lambda z1_arg1, z1_arg2: stack([arctan2(z1_arg1, z1_arg2), 0.0, 0.0]),
            lambda z1_arg1, z1_arg2: stack([-arctan2(z1_arg1, z1_arg2), np.pi, 0.0]),
            z1_arg1,
            z1_arg2,
        )

    if y_arg > 0:
        z1 = arctan2(z1_arg1, z1_arg2)
        y = z2 = 0.0
    else:
        z1 = -arctan2(z1_arg1, z1_arg2)
        y = np.pi
        z2 = 0.0
    return stack([z1, y, z2])


def _regular_quat_to_zyz(qw, qx, qy, qz, y_arg):
    """Compute the ZYZ angles for the regular case (qx != 0 or qy != 0)"""
    z1_arg1 = 2 * (qy * qz - qw * qx)
    z1_arg2 = 2 * (qx * qz + qw * qy)
    z1 = arctan2(z1_arg1, z1_arg2)

    y = arccos(y_arg)

    z2_arg1 = 2 * (qy * qz + qw * qx)
    z2_arg2 = 2 * (qw * qy - qx * qz)
    z2 = arctan2(z2_arg1, z2_arg2)

    return stack([z1, y, z2])


def _fuse(angles_1, angles_2, abstract_jax=False):
    """Perform fusion of two angle sets. Separated out so we can do JIT with conditionals."""
    # Compute the product of the quaternions
    qw, qx, qy, qz = _quaternion_product(_zyz_to_quat(angles_1), _zyz_to_quat(angles_2))

    # Convert the product back into the angles fed to Rot
    y_arg = 1 - 2 * (qx**2 + qy**2)

    # Require special treatment of the case qx = qy = 0. Note that we have to check
    # for "greater than" as well, because of imprecisions
    if abstract_jax:
        from jax.lax import cond

        return cond(
            math_abs(y_arg) >= 1,
            partial(_singular_quat_to_zyz, abstract_jax=True),
            _regular_quat_to_zyz,
            qw,
            qx,
            qy,
            qz,
            y_arg,
        )

    # Require special treatment of the case qx = qy = 0
    if abs(y_arg) >= 1:  # Have to check for "greater than" as well, because of imprecisions
        return _singular_quat_to_zyz(qw, qx, qy, qz, y_arg)
    return _regular_quat_to_zyz(qw, qx, qy, qz, y_arg)


def _no_fuse(angles_1, angles_2):
    """Special case: do not perform fusion when both Y angles are zero:
        Rot(a, 0, b) Rot(c, 0, d) = Rot(a + b + c + d, 0, 0)
    The quaternion math itself will fail in this case without a conditional.
    """
    return stack([angles_1[0] + angles_1[2] + angles_2[0] + angles_2[2], 0.0, 0.0])


def fuse_rot_angles(angles_1, angles_2):
    """Computed the set of rotation angles that is obtained when composing
    two ``qml.Rot`` operations.

    The ``qml.Rot`` operation represents the most general single-qubit operation.
    Two such operations can be fused into a new operation, however the angular dependence
    is non-trivial.

    Args:
        angles_1 (float): A set of three angles for the first ``qml.Rot`` operation.
        angles_2 (float): A set of three angles for the second ``qml.Rot`` operation.

    Returns:
        array[float]: Rotation angles for a single ``qml.Rot`` operation that
        implements the same operation as the two sets of input angles.
    """

    # Check if we are tracing; if so, use the special conditionals
    if is_abstract(angles_1) or is_abstract(angles_2):
        interface = get_interface(angles_1, angles_2)

        # TODO: implement something similar for torch and tensorflow interfaces
        # If the interface is JAX, use jax.lax.cond so that we can jit even with conditionals
        if interface == "jax":
            from jax.lax import cond

            return cond(
                allclose(angles_1[1], 0.0) * allclose(angles_2[1], 0.0),
                _no_fuse,
                partial(_fuse, abstract_jax=True),
                angles_1,
                angles_2,
            )

    # For other interfaces where we would not be jitting or tracing, we can simply check
    # if we are dealing with the special case of Rot(a, 0, b) Rot(c, 0, d).
    if allclose(angles_1[1], 0.0) and allclose(angles_2[1], 0.0):
        return _no_fuse(angles_1, angles_2)

    return _fuse(angles_1, angles_2)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the QuantumPhaseEstimation template.
"""
# pylint: disable=too-many-arguments,arguments-differ
import pennylane as qml
from pennylane.queuing import QueuingManager
from pennylane.operation import AnyWires, Operation, Operator


class QuantumPhaseEstimation(Operation):
    r"""Performs the
    `quantum phase estimation <https://en.wikipedia.org/wiki/Quantum_phase_estimation_algorithm>`__
    circuit.

    Given a unitary matrix :math:`U`, this template applies the circuit for quantum phase
    estimation. The unitary is applied to the qubits specified by ``target_wires`` and :math:`n`
    qubits are used for phase estimation as specified by ``estimation_wires``.

    .. figure:: ../../_static/templates/subroutines/qpe.svg
        :align: center
        :width: 60%
        :target: javascript:void(0);

    This circuit can be used to perform the standard quantum phase estimation algorithm, consisting
    of the following steps:

    #. Prepare ``target_wires`` in a given state. If ``target_wires`` are prepared in an eigenstate
       of :math:`U` that has corresponding eigenvalue :math:`e^{2 \pi i \theta}` with phase
       :math:`\theta \in [0, 1)`, this algorithm will measure :math:`\theta`. Other input states can
       be prepared more generally.
    #. Apply the ``QuantumPhaseEstimation`` circuit.
    #. Measure ``estimation_wires`` using :func:`~.probs`, giving a probability distribution over
       measurement outcomes in the computational basis.
    #. Find the index of the largest value in the probability distribution and divide that number by
       :math:`2^{n}`. This number will be an estimate of :math:`\theta` with an error that decreases
       exponentially with the number of qubits :math:`n`.

    Note that if :math:`\theta \in (-1, 0]`, we can estimate the phase by again finding the index
    :math:`i` found in step 4 and calculating :math:`\theta \approx \frac{1 - i}{2^{n}}`. The
    usage details below give an example of this case.

    Args:
        unitary (array or Operator): the phase estimation unitary, specified as a matrix or an
            :class:`~.Operator`
        target_wires (Union[Wires, Sequence[int], or int]): the target wires to apply the unitary.
            If the unitary is specified as an operator, the target wires should already have been
            defined as part of the operator. In this case, target_wires should not be specified.
        estimation_wires (Union[Wires, Sequence[int], or int]): the wires to be used for phase
            estimation

    Raises:
        QuantumFunctionError: if the ``target_wires`` and ``estimation_wires`` share a common
            element, or if ``target_wires`` are specified for an operator unitary.

    .. details::
        :title: Usage Details

        Consider the matrix corresponding to a rotation from an :class:`~.RX` gate:

        .. code-block:: python

            import pennylane as qml
            from pennylane.templates import QuantumPhaseEstimation
            from pennylane import numpy as np

            phase = 5
            target_wires = [0]
            unitary = qml.RX(phase, wires=0).matrix()

        The ``phase`` parameter can be estimated using ``QuantumPhaseEstimation``. An example is
        shown below using a register of five phase-estimation qubits:

        .. code-block:: python

            n_estimation_wires = 5
            estimation_wires = range(1, n_estimation_wires + 1)

            dev = qml.device("default.qubit", wires=n_estimation_wires + 1)

            @qml.qnode(dev)
            def circuit():
                # Start in the |+> eigenstate of the unitary
                qml.Hadamard(wires=target_wires)

                QuantumPhaseEstimation(
                    unitary,
                    target_wires=target_wires,
                    estimation_wires=estimation_wires,
                )

                return qml.probs(estimation_wires)

            phase_estimated = np.argmax(circuit()) / 2 ** n_estimation_wires

            # Need to rescale phase due to convention of RX gate
            phase_estimated = 4 * np.pi * (1 - phase_estimated)

        We can also perform phase estimation on an operator. Note that since operators are defined
        with target wires, the target wires should not be provided for the QPE.

        .. code-block:: python


            # use the product to specify compound operators
            unitary = qml.RX(np.pi / 2, wires=[0]) @ qml.CNOT(wires=[0, 1])
            eigenvector = np.array([-1/2, -1/2, 1/2, 1/2])

            n_estimation_wires = 5
            estimation_wires = range(2, n_estimation_wires + 2)
            target_wires = [0, 1]

            dev = qml.device("default.qubit", wires=n_estimation_wires + 2)

            @qml.qnode(dev)
            def circuit():
                qml.QubitStateVector(eigenvector, wires=target_wires)
                QuantumPhaseEstimation(
                    unitary,
                    estimation_wires=estimation_wires,
                )
                return qml.probs(estimation_wires)

            phase_estimated = np.argmax(circuit()) / 2 ** n_estimation_wires

    """
    num_wires = AnyWires
    grad_method = None

    def __init__(self, unitary, target_wires=None, estimation_wires=None, do_queue=True, id=None):

        if isinstance(unitary, Operator):

            # If the unitary is expressed in terms of operators, do not provide target wires
            if target_wires is not None:
                raise qml.QuantumFunctionError(
                    "The unitary is expressed as an operator, which already has target wires "
                    "defined, do not additionally specify target wires."
                )
            target_wires = unitary.wires

        elif target_wires is None:
            raise qml.QuantumFunctionError(
                "Target wires must be specified if the unitary is expressed as a matrix."
            )

        else:
            unitary = qml.QubitUnitary(unitary, wires=target_wires)

        # Estimation wires are required, but kept as an optional argument so that it can be
        # placed after target_wires for backwards compatibility.
        if estimation_wires is None:
            raise qml.QuantumFunctionError("No estimation wires specified.")

        target_wires = list(target_wires)
        estimation_wires = list(estimation_wires)
        wires = target_wires + estimation_wires

        if any(wire in target_wires for wire in estimation_wires):
            raise qml.QuantumFunctionError(
                "The target wires and estimation wires must not overlap."
            )

        self._hyperparameters = {
            "unitary": unitary,
            "target_wires": target_wires,
            "estimation_wires": estimation_wires,
        }

        super().__init__(wires=wires, do_queue=do_queue, id=id)

    @property
    def target_wires(self):
        """The target wires of the QPE"""
        return self._hyperparameters["target_wires"]

    @property
    def estimation_wires(self):
        """The estimation wires of the QPE"""
        return self._hyperparameters["estimation_wires"]

    # pylint: disable=protected-access
    def map_wires(self, wire_map: dict):
        new_op = super().map_wires(wire_map)
        new_op._hyperparameters["unitary"] = qml.map_wires(
            new_op._hyperparameters["unitary"], wire_map
        )

        new_op._hyperparameters["estimation_wires"] = [
            wire_map.get(wire, wire) for wire in self.estimation_wires
        ]
        new_op._hyperparameters["target_wires"] = [
            wire_map.get(wire, wire) for wire in self.target_wires
        ]

        return new_op

    def queue(self, context=QueuingManager):
        context.update_info(self._hyperparameters["unitary"], owner=self)
        context.append(self, owns=self._hyperparameters["unitary"])
        return self

    @staticmethod
    def compute_decomposition(
        wires, unitary, target_wires, estimation_wires
    ):  # pylint: disable=arguments-differ,unused-argument
        r"""Representation of the QPE circuit as a product of other operators.

        .. math:: O = O_1 O_2 \dots O_n.


        .. seealso:: :meth:`~.QuantumPhaseEstimation.decomposition`.

        Args:
            wires (Any or Iterable[Any]): wires that the QPE circuit acts on
            unitary (Operator): the phase estimation unitary, specified as an operator
            target_wires (Any or Iterable[Any]): the target wires to apply the unitary
            estimation_wires (Any or Iterable[Any]): the wires to be used for phase estimation

        Returns:
            list[.Operator]: decomposition of the operator
        """

        op_list = [qml.Hadamard(w) for w in estimation_wires]
        pow_ops = (qml.pow(unitary, 2**i) for i in range(len(estimation_wires) - 1, -1, -1))
        op_list.extend(qml.ctrl(op, w) for op, w in zip(pow_ops, estimation_wires))
        op_list.append(qml.adjoint(qml.templates.QFT(wires=estimation_wires)))

        return op_list



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the qml.measure measurement.
"""
import uuid
from typing import Generic, TypeVar

import pennylane as qml
import pennylane.numpy as np
from pennylane.operation import Wires

from .measurements import MeasurementProcess, MidMeasure


def measure(wires):  # TODO: Change name to mid_measure
    """Perform a mid-circuit measurement in the computational basis on the
    supplied qubit.

    Measurement outcomes can be obtained and used to conditionally apply
    operations.

    If a device doesn't support mid-circuit measurements natively, then the
    QNode will apply the :func:`defer_measurements` transform.

    **Example:**

    .. code-block:: python3

        dev = qml.device("default.qubit", wires=2)

        @qml.qnode(dev)
        def func(x, y):
            qml.RY(x, wires=0)
            qml.CNOT(wires=[0, 1])
            m_0 = qml.measure(1)

            qml.cond(m_0, qml.RY)(y, wires=0)
            return qml.probs(wires=[0])

    Executing this QNode:

    >>> pars = np.array([0.643, 0.246], requires_grad=True)
    >>> func(*pars)
    tensor([0.90165331, 0.09834669], requires_grad=True)

    Mid circuit measurements can be manipulated using the following dunder methods
    `+`, `-`, `*`, `/`, `~` (not), `&` (and), `|` (or), `==`, `<=`, `>=`, `<`, `>`. With other mid-circuit measurements or scalars.

    Note:
        python `not`, `and`, `or`, do not work since these do not have dunder methods. Instead use `~`, `&`, `|`.

    Args:
        wires (Wires): The wire of the qubit the measurement process applies to.

    Returns:
        MidMeasureMP: measurement process instance

    Raises:
        QuantumFunctionError: if multiple wires were specified
    """
    wire = Wires(wires)
    if len(wire) > 1:
        raise qml.QuantumFunctionError(
            "Only a single qubit can be measured in the middle of the circuit"
        )

    # Create a UUID and a map between MP and MV to support serialization
    measurement_id = str(uuid.uuid4())[:8]
    MidMeasureMP(wires=wire, id=measurement_id)
    return MeasurementValue([measurement_id], processing_fn=lambda v: v)


T = TypeVar("T")


class MidMeasureMP(MeasurementProcess):
    """Mid-circuit measurement.

    Please refer to :func:`measure` for detailed documentation.

    Args:
        obs (.Observable): The observable that is to be measured as part of the
            measurement process. Not all measurement processes require observables (for
            example ``Probability``); this argument is optional.
        wires (.Wires): The wires the measurement process applies to.
            This can only be specified if an observable was not provided.
        eigvals (array): A flat array representing the eigenvalues of the measurement.
            This can only be specified if an observable was not provided.
        id (str): custom label given to a measurement instance, can be useful for some applications
            where the instance has to be identified
    """

    @property
    def return_type(self):
        return MidMeasure

    @property
    def _queue_category(self):
        return "_ops"


class MeasurementValueError(ValueError):
    """Error raised when an unknown measurement value is being used."""


class MeasurementValue(Generic[T]):
    """A class representing unknown measurement outcomes in the qubit model.

    Measurements on a single qubit in the computational basis are assumed.

    Args:
        measurement_ids (list[str]): The id of the measurement that this object depends on.
        processing_fn (callable): A lazily transformation applied to the measurement values.
    """

    def __init__(self, measurement_ids, processing_fn):
        self.measurement_ids = measurement_ids
        self.processing_fn = processing_fn

    def _items(self):
        """A generator representing all the possible outcomes of the MeasurementValue."""
        for i in range(2 ** len(self.measurement_ids)):
            branch = tuple(int(b) for b in np.binary_repr(i, width=len(self.measurement_ids)))
            yield branch, self.processing_fn(*branch)

    @property
    def branches(self):
        """A dictionary representing all possible outcomes of the MeasurementValue."""
        ret_dict = {}
        for i in range(2 ** len(self.measurement_ids)):
            branch = tuple(int(b) for b in np.binary_repr(i, width=len(self.measurement_ids)))
            ret_dict[branch] = self.processing_fn(*branch)
        return ret_dict

    def _transform_bin_op(self, base_bin, other):
        """Helper function for defining dunder binary operations."""
        if isinstance(other, MeasurementValue):
            # pylint: disable=protected-access
            return self._merge(other)._apply(lambda t: base_bin(t[0], t[1]))
        # if `other` is not a MeasurementValue then apply it to each branch
        return self._apply(lambda v: base_bin(v, other))

    def __invert__(self):
        """Return a copy of the measurement value with an inverted control
        value."""
        return self._apply(lambda v: not v)

    def __eq__(self, other):
        return self._transform_bin_op(lambda a, b: a == b, other)

    def __ne__(self, other):
        return self._transform_bin_op(lambda a, b: a != b, other)

    def __add__(self, other):
        return self._transform_bin_op(lambda a, b: a + b, other)

    def __radd__(self, other):
        return self._apply(lambda v: other + v)

    def __sub__(self, other):
        return self._transform_bin_op(lambda a, b: a - b, other)

    def __rsub__(self, other):
        return self._apply(lambda v: other - v)

    def __mul__(self, other):
        return self._transform_bin_op(lambda a, b: a * b, other)

    def __rmul__(self, other):
        return self._apply(lambda v: other * v)

    def __truediv__(self, other):
        return self._transform_bin_op(lambda a, b: a / b, other)

    def __rtruediv__(self, other):
        return self._apply(lambda v: other / v)

    def __lt__(self, other):
        return self._transform_bin_op(lambda a, b: a < b, other)

    def __le__(self, other):
        return self._transform_bin_op(lambda a, b: a <= b, other)

    def __gt__(self, other):
        return self._transform_bin_op(lambda a, b: a > b, other)

    def __ge__(self, other):
        return self._transform_bin_op(lambda a, b: a >= b, other)

    def __and__(self, other):
        return self._transform_bin_op(lambda a, b: a and b, other)

    def __or__(self, other):
        return self._transform_bin_op(lambda a, b: a or b, other)

    def _apply(self, fn):
        """Apply a post computation to this measurement"""
        return MeasurementValue(self.measurement_ids, lambda *x: fn(self.processing_fn(*x)))

    def _merge(self, other: "MeasurementValue"):
        """Merge two measurement values"""

        # create a new merged list with no duplicates and in lexical ordering
        merged_measurement_ids = list(set(self.measurement_ids).union(set(other.measurement_ids)))
        merged_measurement_ids.sort()

        # create a new function that selects the correct indices for each sub function
        def merged_fn(*x):
            sub_args_1 = (
                x[i] for i in [merged_measurement_ids.index(m) for m in self.measurement_ids]
            )
            out_1 = self.processing_fn(*sub_args_1)

            sub_args_2 = (
                x[i] for i in [merged_measurement_ids.index(m) for m in other.measurement_ids]
            )
            out_2 = other.processing_fn(*sub_args_2)

            return out_1, out_2

        return MeasurementValue(merged_measurement_ids, merged_fn)

    def __getitem__(self, i):
        branch = tuple(int(b) for b in np.binary_repr(i, width=len(self.measurement_ids)))
        return self.processing_fn(*branch)

    def __str__(self):
        lines = []
        for i in range(2 ** (len(self.measurement_ids))):
            branch = tuple(int(b) for b in np.binary_repr(i, width=len(self.measurement_ids)))
            id_branch_mapping = [
                f"{self.measurement_ids[j]}={branch[j]}" for j in range(len(branch))
            ]
            lines.append(
                "if " + ",".join(id_branch_mapping) + " => " + str(self.processing_fn(*branch))
            )
        return "\n".join(lines)



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the qml.is_hermitian function.
"""
import pennylane as qml
from pennylane.operation import Operator


def is_hermitian(op: Operator):
    r"""Check if the operation is hermitian.

    A hermitian matrix is a complex square matrix that is equal to its own adjoint

    .. math:: O^\dagger = O

    Args:
        op (~.operation.Operator): the operator to check against

    Returns:
        bool: True if the operation is hermitian, False otherwise

    .. note::
        This check might be expensive for large operators.

    **Example**

    >>> op = qml.PauliX(wires=0)
    >>> qml.is_hermitian(op)
    True
    >>> op2 = qml.RX(0.54, wires=0)
    >>> qml.is_hermitian(op2)
    False
    """
    if op.is_hermitian is True:
        return True
    return qml.math.allclose(qml.matrix(op), qml.matrix(qml.adjoint(op)))



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the Identity operation that is common to both
cv and qubit computing paradigms in PennyLane.
"""
from functools import lru_cache

from scipy import sparse

import pennylane as qml
from pennylane.operation import AnyWires, CVObservable, Operation


class Identity(CVObservable, Operation):
    r"""pennylane.Identity(wires)
    The identity observable :math:`\I`.

    The expectation of this observable

    .. math::
        E[\I] = \text{Tr}(\I \rho)

    corresponds to the trace of the quantum state, which in exact
    simulators should always be equal to 1.
    """
    num_params = 0
    num_wires = AnyWires
    """int: Number of wires that the operator acts on."""

    grad_method = None
    """Gradient computation method."""

    _queue_category = "_ops"

    ev_order = 1

    def __init__(self, *params, wires=None, do_queue=True, id=None):
        super().__init__(*params, wires=wires, do_queue=do_queue, id=id)
        self._hyperparameters = {"n_wires": len(self.wires)}
        self._pauli_rep = qml.pauli.PauliSentence({qml.pauli.PauliWord({}): 1.0})

    def label(self, decimals=None, base_label=None, cache=None):
        return base_label or "I"

    @staticmethod
    def compute_eigvals(n_wires=1):  # pylint: disable=arguments-differ
        r"""Eigenvalues of the operator in the computational basis (static method).

        If :attr:`diagonalizing_gates` are specified and implement a unitary :math:`U^{\dagger}`,
        the operator can be reconstructed as

        .. math:: O = U \Sigma U^{\dagger},

        where :math:`\Sigma` is the diagonal matrix containing the eigenvalues.

        Otherwise, no particular order for the eigenvalues is guaranteed.

        .. seealso:: :meth:`~.Identity.eigvals`

        Returns:
            array: eigenvalues

        **Example**

        >>> print(qml.Identity.compute_eigvals())
        [ 1 1]
        """
        return qml.math.ones(2**n_wires)

    @staticmethod
    @lru_cache()
    def compute_matrix(n_wires=1):  # pylint: disable=arguments-differ
        r"""Representation of the operator as a canonical matrix in the computational basis (static method).

        The canonical matrix is the textbook matrix representation that does not consider wires.
        Implicitly, this assumes that the wires of the operator correspond to the global wire order.

        .. seealso:: :meth:`~.Identity.matrix`

        Returns:
            ndarray: matrix

        **Example**

        >>> print(qml.Identity.compute_matrix())
        [[1. 0.]
         [0. 1.]]
        """
        return qml.math.eye(int(2**n_wires))

    @staticmethod
    @lru_cache()
    def compute_sparse_matrix(n_wires=1):  # pylint: disable=arguments-differ
        return sparse.eye(int(2**n_wires), format="csr")

    @staticmethod
    def _heisenberg_rep(p):
        return qml.math.array([1, 0, 0])

    @staticmethod
    def compute_diagonalizing_gates(
        wires, n_wires=1
    ):  # pylint: disable=arguments-differ,unused-argument
        r"""Sequence of gates that diagonalize the operator in the computational basis (static method).

        Given the eigendecomposition :math:`O = U \Sigma U^{\dagger}` where
        :math:`\Sigma` is a diagonal matrix containing the eigenvalues,
        the sequence of diagonalizing gates implements the unitary :math:`U^{\dagger}`.

        The diagonalizing gates rotate the state into the eigenbasis
        of the operator.

        .. seealso:: :meth:`~.Identity.diagonalizing_gates`.

        Args:
            wires (Iterable[Any], Wires): wires that the operator acts on

        Returns:
            list[.Operator]: list of diagonalizing gates

        **Example**

        >>> qml.Identity.compute_diagonalizing_gates(wires=[0])
        []
        """
        return []

    @staticmethod
    def compute_decomposition(wires, n_wires=1):  # pylint:disable=arguments-differ,unused-argument
        r"""Representation of the operator as a product of other operators (static method).

        .. math:: O = O_1 O_2 \dots O_n.

        .. seealso:: :meth:`~.Identity.decomposition`.

        Args:
            wires (Any, Wires): A single wire that the operator acts on.

        Returns:
            list[Operator]: decomposition into lower level operations

        **Example:**

        >>> qml.Identity.compute_decomposition(wires=0)
        []

        """
        return []

    @staticmethod
    def identity_op(*params):
        """Alias for matrix representation of the identity operator."""
        return Identity.compute_matrix(*params)

    def adjoint(self):
        return Identity(wires=self.wires)

    def pow(self, _):
        return [Identity(wires=self.wires)]



# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class and functions for activating, deactivating and checking the new return types system
"""
# pylint: disable=global-statement
__activated = False


def enable_return():
    """Function that turns on the experimental return type system that prefers the use of sequences over arrays.

    The new system guarantees that a sequence (e.g., list or tuple) is returned based on the ``return`` statement of the
    quantum function. This system avoids the creation of ragged arrays, where multiple measurements are stacked
    together.

    **Example**

    The following example shows that for multiple measurements the current PennyLane system is creating a ragged tensor.

    .. code-block:: python

        dev = qml.device("default.qubit", wires=2)

        def circuit(x):
              qml.Hadamard(wires=[0])
              qml.CRX(x, wires=[0, 1])
              return qml.probs(wires=[0]), qml.vn_entropy(wires=[0]), qml.probs(wires=1), qml.expval(qml.PauliZ(wires=1))

        qnode = qml.QNode(circuit, dev)

    >>> qnode(0.5)
    tensor([0.5       , 0.5       , 0.08014815, 0.96939564, 0.03060436,
        0.93879128], requires_grad=True)

    when you activate the new return type the result is simply a tuple containing each measurement.

    .. code-block:: python

        qml.enable_return()

        dev = qml.device("default.qubit", wires=2)

        def circuit(x):
              qml.Hadamard(wires=[0])
              qml.CRX(x, wires=[0, 1])
              return qml.probs(wires=[0]), qml.vn_entropy(wires=[0]), qml.probs(wires=1), qml.expval(qml.PauliZ(wires=1))

        qnode = qml.QNode(circuit, dev)

    >>> qnode(0.5)
    (tensor([0.5, 0.5], requires_grad=True), tensor(0.08014815, requires_grad=True), tensor([0.96939564, 0.03060436], requires_grad=True), tensor(0.93879128, requires_grad=True))

    .. note::

        This is an experimental feature and may not support every feature in PennyLane. The list of supported features
        from PennyLane include:

        * :func:`~.pennylane.execute`
        * Gradient transforms

          #. :func:`~.pennylane.gradients.param_shift`;
          #. :func:`~.pennylane.gradients.finite_diff`;
          #. :class:`~.pennylane.gradients.hessian_transform`;
          #. :func:`~.pennylane.gradients.param_shift_hessian`.

        * Interfaces

          #. Autograd;
          #. TensorFlow;
          #. JAX (without jitting);

        * PennyLane optimizers
        * :meth:`~.pennylane.tape.QuantumTape.shape`

    Note that this is an experimental feature and may not support every feature in PennyLane. See the ``Usage Details``
    section for more details.

    .. details::
        :title: Usage Details

        **Gotcha: Autograd and TensorFlow can only compute gradients of tensor-valued functions**

        Autograd and TensorFlow only allow differentiating functions that have array or tensor outputs. QNodes that
        have multiple measurements may output other sequences with the new return types may cause errors with Autograd
        or TensorFlow.

        This issue can be overcome by stacking the QNode results before computing derivatives:

        .. code-block:: python

            qml.enable_return()

            a = np.array(0.1, requires_grad=True)
            b = np.array(0.2, requires_grad=True)

            dev = qml.device("lightning.qubit", wires=2)

            @qml.qnode(dev, diff_method="parameter-shift")
            def circuit(a, b):
                qml.RY(a, wires=0)
                qml.RX(b, wires=1)
                qml.CNOT(wires=[0, 1])
                return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliY(1))

            def cost(x, y):
                return qml.numpy.hstack(circuit(x, y))

        >>> qml.jacobian(cost)(a, b)
        (array([-0.09983342,  0.01983384]), array([-5.54649074e-19, -9.75170327e-01]))

        If no stacking is performed, Autograd raises the following error:

        .. code-block:: python

            TypeError: 'ArrayVSpace' object cannot be interpreted as an integer

        The solution with TensorFlow is similar with the difference that stacking happens within the
        ``tf.GradientTape()`` context:

        .. code-block:: python

            a = tf.Variable(0.1, dtype=tf.float64)
            b = tf.Variable(0.2, dtype=tf.float64)

            dev = qml.device("default.qubit", wires=2)

            @qml.qnode(dev, diff_method="parameter-shift", interface="tf")
            def circuit(a, b):
                qml.RY(a, wires=0)
                qml.RX(b, wires=1)
                qml.CNOT(wires=[0, 1])
                return qml.expval(qml.PauliZ(0)), qml.expval(qml.PauliY(1))

            with tf.GradientTape() as tape:
                res = circuit(a, b)
                res = tf.stack(res)

            assert circuit.qtape.trainable_params == [0, 1]

            tape.jacobian(res, [a, b])

        If the measurements do not have the same shape then you need to use TF hstack:

        .. code-block:: python

            a = tf.Variable(0.1, dtype=tf.float64)
            b = tf.Variable(0.2, dtype=tf.float64)

            dev = qml.device("default.qubit", wires=2)

            @qml.qnode(dev, diff_method="parameter-shift", interface="tf")
            def circuit(a, b):
                qml.RY(a, wires=0)
                qml.RX(b, wires=1)
                qml.CNOT(wires=[0, 1])
                return qml.expval(qml.PauliZ(0)), qml.probs(wires=[0, 1])

            with tf.GradientTape() as tape:
                res = circuit(a, b)
                res = tf.experimental.numpy.hstack(res)

            assert circuit.qtape.trainable_params == [0, 1]

            tape.jacobian(res, [a, b])

        If no stacking is performed, TensorFlow raises the following error:

        .. code-block:: python

            AttributeError: 'tuple' object has no attribute 'shape'

        **JAX interface upgrades: higher-order derivatives and mixing measurements**

        Higher-order derivatives can now be computed with the JAX interface:

        .. code-block:: python

            import jax

            qml.enable_return()

            dev = qml.device("lightning.qubit", wires=2)

            par_0 = jax.numpy.array(0.1)
            par_1 = jax.numpy.array(0.2)

            @qml.qnode(dev, interface="jax", diff_method="parameter-shift", max_diff=2)
            def circuit(x, y):
                qml.RX(x, wires=[0])
                qml.RY(y, wires=[1])
                qml.CNOT(wires=[0, 1])
                return qml.expval(qml.PauliZ(0) @ qml.PauliX(1))

        >>> jax.hessian(circuit, argnums=[0, 1])(par_0, par_1)
        ((Array(-0.19767681, dtype=float32, weak_type=True),
          Array(-0.09784342, dtype=float32, weak_type=True)),
         (Array(-0.09784339, dtype=float32, weak_type=True),
          Array(-0.19767687, dtype=float32, weak_type=True)))

        The new return types system also unlocks the use of ``probs`` mixed with different measurements with JAX:

        .. code-block:: python

            import jax

            qml.enable_return()

            dev = qml.device("default.qubit", wires=2)
            qml.enable_return()

            @qml.qnode(dev, interface="jax")
            def circuit(a):
              qml.RX(a[0], wires=0)
              qml.CNOT(wires=(0, 1))
              qml.RY(a[1], wires=1)
              qml.RZ(a[2], wires=1)
              return qml.expval(qml.PauliZ(wires=0)), qml.probs(wires=[0, 1])

            x = jax.numpy.array([0.1, 0.2, 0.3])

        >>> jax.jacobian(circuit)(x)
        (Array([-9.9833414e-02, -7.4505806e-09,  6.9285655e-10], dtype=float32),
         Array([[-4.9419206e-02, -9.9086545e-02,  3.4938008e-09],
                      [-4.9750542e-04,  9.9086538e-02,  1.2768372e-10],
                      [ 4.9750548e-04,  2.4812977e-04,  4.8371929e-13],
                      [ 4.9419202e-02, -2.4812980e-04,  2.6696912e-11]],            dtype=float32))

        where before the following error was raised:

        .. code-block:: python

            ValueError: All input arrays must have the same shape.

        The new return types system also unlocks the use of shot vectors with all the previous features. For example you
        can take the second derivative and multiple measurement with with JAX:

        .. code-block:: python

            import jax

            qml.enable_return()

            dev = qml.device("default.qubit", wires=2, shots=(1, 10000))

            params = jax.numpy.array([0.1, 0.2])

            @qml.qnode(dev, interface="jax", diff_method="parameter-shift", max_diff=2)
            def circuit(x):
                qml.RX(x[0], wires=[0])
                qml.RY(x[1], wires=[1])
                qml.CNOT(wires=[0, 1])
                return qml.var(qml.PauliZ(0) @ qml.PauliX(1)), qml.probs(wires=[0])

        >>> jax.hessian(circuit)(params)
        ((Array([[ 0.,  0.],
                        [ 2., -3.]], dtype=float32),
          Array([[[-0.5,  0. ],
                       [ 0. ,  0. ]],
                      [[ 0.5,  0. ],
                       [ 0. ,  0. ]]], dtype=float32)),
         (Array([[ 0.07677898,  0.0563341 ],
                       [ 0.07238522, -1.830669  ]], dtype=float32),
          Array([[[-4.9707499e-01,  2.9999996e-04],
                        [-6.2500127e-04,  1.2500001e-04]],
                       [[ 4.9707499e-01, -2.9999996e-04],
                        [ 6.2500127e-04, -1.2500001e-04]]], dtype=float32)))

    .. details::
        :title: Usage Details for Autograd

        Autograd only allows differentiating functions that have array or tensor outputs. QNodes that have
        multiple measurements may output tuples and cause errors with Autograd. Similarly, the outputs of
        ``qml.grad`` or ``qml.jacobian`` may be tuples if there are multiple trainable arguments and
        cause errors with higher-order derivatives.

        These issues can be overcome by stacking the QNode results or gradients before computing derivatives.

        **Examples for shot vectors**

        Example for first-order derivatives with a single return:

        .. code-block:: python

            def func(a, b):
                qml.RY(a, wires=0)
                qml.RX(b, wires=1)
                qml.CNOT(wires=[0, 1])

            dev = qml.device("default.qubit", wires=2, shots=[100, 200, 300])

            @qml.qnode(dev, diff_method="parameter-shift")
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0))

            def cost(a, b):
                res = circuit(a, b)
                return qml.math.stack(res)

        >>> a = np.array(0.4, requires_grad=True)
        >>> b = np.array(0.6, requires_grad=True)
        >>> cost(a, b)
        [0.94       0.93       0.89333333]
        >>> qml.jacobian(cost)(a, b)
        (array([-0.44, -0.38, -0.38]), array([-0.01      , -0.01      , -0.00666667]))

        Example for first-order derivatives with multiple returns:

        .. code-block:: python

            @qml.qnode(dev, diff_method="parameter-shift")
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0)), qml.probs([0, 1])

            def cost(a, b):
                res = circuit(a, b)
                return qml.math.stack([qml.math.hstack(r) for r in res])

        >>> cost(a, b)
        [[0.92       0.91       0.05       0.03       0.01      ]
         [0.95       0.9        0.075      0.         0.025     ]
         [0.95333333 0.89666667 0.08       0.         0.02333333]]
        >>> qml.jacobian(cost)(a, b)
        (array([[-0.41  , -0.16  , -0.045 ,  0.01  ,  0.195 ],
                [-0.325 , -0.1325, -0.03  ,  0.01  ,  0.1525],
                [-0.39  , -0.175 , -0.02  ,  0.02  ,  0.175 ]]),
         array([[-0.01      , -0.26      ,  0.255     ,  0.02      , -0.015     ],
                [ 0.01      , -0.2525    ,  0.2575    ,  0.0075    , -0.0125    ],
                [-0.02      , -0.26666667,  0.25666667,  0.01666667, -0.00666667]]))

        Example for second-order derivatives with a single return:

        .. code-block:: python

            @qml.qnode(dev, diff_method="parameter-shift", max_diff=2)
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1))

            def cost(a, b):
                def cost2(a, b):
                    res = circuit(a, b)
                    return qml.math.stack(res)

                return qml.math.stack(qml.jacobian(cost2)(a, b))

        >>> cost(a, b)
        [[ 0.05       -0.01        0.00666667]
         [-0.51       -0.48       -0.62      ]]
        >>> qml.jacobian(cost)(a, b)
        (array([[-0.03      , -0.02      , -0.00333333],
                [ 0.01      ,  0.025     , -0.02      ]]),
         array([[ 0.005     ,  0.03      , -0.02166667],
                [-0.85      , -0.83      , -0.81333333]]))

        Example for second-order derivatives with multiple returns:

        .. code-block:: python

            @qml.qnode(dev, diff_method="parameter-shift", max_diff=2)
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1)), qml.probs([0, 1])

            def cost(a, b):
                def cost2(a, b):
                    res = circuit(a, b)
                    return qml.math.stack([qml.math.hstack(r) for r in res])

                return qml.math.stack(qml.jacobian(cost2)(a, b))

        >>> cost(a, b)
        [[[-0.05       -0.2         0.          0.025       0.175     ]
          [ 0.02       -0.17       -0.0275      0.0175      0.18      ]
          [-0.00666667 -0.19166667 -0.01333333  0.01666667  0.18833333]]
         [[-0.52       -0.25        0.265      -0.005      -0.01      ]
          [-0.54       -0.25        0.2625      0.0075     -0.02      ]
          [-0.57666667 -0.275       0.28166667  0.00666667 -0.01333333]]]
        >>> qml.jacobian(cost)(a, b)
        (array([[[ 0.02      , -0.38      , -0.055     ,  0.045     ,  0.39      ],
                 [ 0.03      , -0.4275    , -0.04      ,  0.025     ,  0.4425    ],
                 [-0.00666667, -0.42333333, -0.025     ,  0.02833333,  0.42      ]],
                [[ 0.        ,  0.0625    , -0.065     ,  0.065     , -0.0625    ],
                 [ 0.015     ,  0.06875   , -0.05      ,  0.0425    , -0.06125   ],
                 [-0.03      ,  0.03833333, -0.04583333,  0.06083333, -0.05333333]]]),
         array([[[-0.06      ,  0.0325    , -0.0525    ,  0.0825    , -0.0625    ],
                 [ 0.02      ,  0.03      , -0.035     ,  0.025     , -0.02      ],
                 [ 0.00833333,  0.06166667, -0.05666667,  0.0525    , -0.0575    ]],
                [[-0.85      , -0.415     ,  0.425     ,  0.        , -0.01      ],
                 [-0.78      , -0.3775    ,  0.375     ,  0.015     , -0.0125    ],
                 [-0.80666667, -0.38      ,  0.38666667,  0.01666667, -0.02333333]]]))

    .. details::
        :title: Usage Details for TensorFlow

        TensorFlow only allows differentiating functions that have array or tensor outputs. QNodes that have
        multiple measurements may output tuples and cause errors with TensorFlow. Similarly, the outputs of
        ``tape.gradient`` or ``tape.jacobian`` may be tuples if there are multiple trainable arguments and
        cause errors with higher-order derivatives.

        These issues can be overcome by stacking the QNode results or gradients before computing derivatives.

        **Examples for shot vectors**

        Example for first-order derivatives with a single return:

        .. code-block:: python

            def func(a, b):
                qml.RY(a, wires=0)
                qml.RX(b, wires=1)
                qml.CNOT(wires=[0, 1])

            dev = qml.device("default.qubit", wires=2, shots=[100, 200, 300])

            @qml.qnode(dev, diff_method="parameter-shift", interface="tf")
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0))

        >>> a = tf.Variable(0.4, dtype=tf.float64, trainable=True)
        >>> b = tf.Variable(0.6, dtype=tf.float64, trainable=True)
        >>> with tf.GradientTape() as tape:
        ...     res = circuit(a, b)
        ...     res = tf.stack(res)
        ...
        >>> res
        tf.Tensor([0.92 0.92 0.94], shape=(3,), dtype=float64)
        >>> tape.jacobian(res, (a, b))
        (<tf.Tensor: shape=(3,), dtype=float64, numpy=array([-0.31      , -0.385     , -0.32666667])>,
         <tf.Tensor: shape=(3,), dtype=float64, numpy=array([-0.02      ,  0.02      , -0.00333333])>)

        Example for first-order derivatives with multiple returns:

        .. code-block:: python

            @qml.qnode(dev, diff_method="parameter-shift", interface="tf")
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0)), qml.probs([0, 1])

        >>> with tf.GradientTape() as tape:
        ...     res = circuit(a, b)
        ...     res = tf.stack([tf.experimental.numpy.hstack(r) for r in res])
        ...
        >>> res
        tf.Tensor(
        [[0.96       0.93       0.05       0.         0.02      ]
         [0.9        0.87       0.08       0.01       0.04      ]
         [0.96       0.86666667 0.11333333 0.         0.02      ]], shape=(3, 5), dtype=float64)
        >>> tape.jacobian(res, (a, b))
        (<tf.Tensor: shape=(3, 5), dtype=float64, numpy=
         array([[-0.3       , -0.145     , -0.005     ,  0.03      ,  0.12      ],
                [-0.395     , -0.19      , -0.0075    ,  0.01      ,  0.1875    ],
                [-0.39333333, -0.17833333, -0.01833333,  0.00666667,  0.19      ]])>,
         <tf.Tensor: shape=(3, 5), dtype=float64, numpy=
         array([[-0.03      , -0.21      ,  0.195     ,  0.025     , -0.01      ],
                [ 0.005     , -0.285     ,  0.2875    ,  0.0075    , -0.01      ],
                [-0.01333333, -0.27      ,  0.26333333,  0.015     , -0.00833333]])>)

        Example for second-order derivatives with a single return:

        .. code-block:: python

            @qml.qnode(dev, diff_method="parameter-shift", max_diff=2, interface="tf")
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1))

        >>> with tf.GradientTape() as tape1:
        ...     with tf.GradientTape(persistent=True) as tape2:
        ...         res = circuit(a, b)
        ...         res = tf.stack(res)
        ...
        ...    jac = tape2.jacobian(res, (a, b), experimental_use_pfor=False)
        ...    jac = tf.stack(jac)
        ...
        >>> jac
        tf.Tensor(
        [[-0.02       -0.005       0.00333333]
         [-0.47       -0.64       -0.58      ]], shape=(2, 3), dtype=float64)
        >>> tape1.jacobian(jac, (a, b))
        (<tf.Tensor: shape=(2, 3), dtype=float64, numpy=
         array([[ 0.00000000e+00,  2.50000000e-02, -1.66666667e-02],
                [-1.50000000e-02, -4.00000000e-02,  2.77555756e-17]])>,
         <tf.Tensor: shape=(2, 3), dtype=float64, numpy=
         array([[-0.015, -0.04 ,  0.   ],
                [-0.84 , -0.825, -0.84 ]])>)

        Example for second-order derivatives with multiple returns:

        .. code-block:: python

            @qml.qnode(dev, diff_method="parameter-shift", max_diff=2, interface="tf")
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1)), qml.probs([0, 1])

        >>> with tf.GradientTape() as tape1:
        ...     with tf.GradientTape(persistent=True) as tape2:
        ...         res = circuit(a, b)
        ...         res = tf.stack([tf.experimental.numpy.hstack(r) for r in res])
        ...
        ...     jac = tape2.jacobian(res, (a, b), experimental_use_pfor=False)
        ...     jac = tf.stack(jac)
        ...
        >>> jac
        tf.Tensor(
        [[[ 0.01       -0.13       -0.03        0.025       0.135     ]
          [-0.045      -0.195      -0.0075      0.03        0.1725    ]
          [-0.00666667 -0.17       -0.01833333  0.02166667  0.16666667]]
         [[-0.61       -0.29        0.315      -0.01       -0.015     ]
          [-0.55       -0.255       0.265       0.01       -0.02      ]
          [-0.65       -0.305       0.31166667  0.01333333 -0.02      ]]], shape=(2, 3, 5), dtype=float64)
        >>> tape1.jacobian(jac, (a, b))
        (<tf.Tensor: shape=(2, 3, 5), dtype=float64, numpy=
         array([[[ 0.07      , -0.415     , -0.055     ,  0.02      ,  0.45      ],
                 [ 0.05      , -0.4       , -0.06      ,  0.035     ,  0.425     ],
                 [ 0.01666667, -0.42166667, -0.04      ,  0.03166667,  0.43      ]],
                [[-0.03      ,  0.0375    , -0.075     ,  0.09      , -0.0525    ],
                 [-0.005     ,  0.0375    , -0.06625   ,  0.06875   , -0.04      ],
                 [-0.04      ,  0.0375    , -0.05166667,  0.07166667, -0.0575    ]]])>,
         <tf.Tensor: shape=(2, 3, 5), dtype=float64, numpy=
         array([[[-0.03      ,  0.0375    , -0.075     ,  0.09      , -0.0525    ],
                 [-0.005     ,  0.0375    , -0.06625   ,  0.06875   , -0.04      ],
                 [-0.04      ,  0.0375    , -0.05166667,  0.07166667, -0.0575    ]],
                [[-0.81      , -0.39      ,  0.385     ,  0.02      , -0.015     ],
                 [-0.77      , -0.375     ,  0.355     ,  0.03      , -0.01      ],
                 [-0.82666667, -0.40166667,  0.4       ,  0.01333333, -0.01166667]]])>)

    .. details::
        :title: Usage Details for PyTorch

        PyTorch supports differentiation of Torch tensors or tuple of Torch tensors. It makes it easy to get the
        Jacobian of functions returning any mix of measurements.

        .. code-block:: python

            def func(a, b):
                qml.RY(a, wires=0)
                qml.RX(b, wires=1)
                qml.CNOT(wires=[0, 1])

            dev = qml.device("default.qubit", wires=2)

            @qml.qnode(dev, diff_method="parameter-shift", interface="torch")
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0)), qml.probs([0, 1])

        >>> a = torch.tensor(0.1, requires_grad=True)
        >>> b = torch.tensor(0.2, requires_grad=True)
        >>> torch.autograd.functional.jacobian(circuit, (a, b))
        ((tensor(-0.0998), tensor(0.)), (tensor([-0.0494, -0.0005,  0.0005,  0.0494]), tensor([-0.0991,  0.0991,  0.0002, -0.0002])))

        An issue arises when one requires higher order differentation with multiple measurements, the Jacobian returns
        a tuple of tuple which is not consider as differentiable by PyTorch. This issue can be overcome by stacking
        the original results:

        .. code-block:: python

            @qml.qnode(dev, diff_method="parameter-shift", interface="torch", max_diff=2)
            def circuit(a, b):
                func(a, b)
                return qml.expval(qml.PauliZ(0)), qml.probs([0, 1])

            a = torch.tensor(0.1, requires_grad=True)
            b = torch.tensor(0.2, requires_grad=True)

            def circuit_stack(a, b):
                return torch.hstack(circuit(a, b))

        >>> jac_fn = lambda a, b: torch.autograd.functional.jacobian(circuit_stack, (a, b), create_graph=True)
        >>> torch.autograd.functional.jacobian(jac_fn, (a, b))
        ((tensor([-0.9950, -0.4925, -0.0050,  0.0050,  0.4925]), tensor([ 0.0000,  0.0050, -0.0050,  0.0050, -0.0050])),
        (tensor([ 0.0000,  0.0050, -0.0050,  0.0050, -0.0050]), tensor([ 0.0000, -0.4888,  0.4888,  0.0012, -0.0012])))

    """

    global __activated
    __activated = True


def disable_return():
    """Function that turns off the new return type system.

    **Example**

    .. code-block:: python

        dev = qml.device("default.qubit", wires=2)

        def circuit(x):
              qml.Hadamard(wires=[0])
              qml.CRX(x, wires=[0, 1])
              return qml.probs(wires=[0]), qml.vn_entropy(wires=[0]), qml.probs(wires=1), qml.expval(qml.PauliZ(wires=1))

        qnode = qml.QNode(circuit, dev)


    >>> qml.enable_return()
    >>> res = qnode(0.5)
    >>> res
    (tensor([0.5, 0.5], requires_grad=True), tensor(0.08014815, requires_grad=True), tensor([0.96939564, 0.03060436], requires_grad=True), tensor(0.93879128, requires_grad=True))
    >>> qml.disable_return()
    >>> res = qnode(0.5)
    >>> res
    tensor([0.5       , 0.5       , 0.08014815, 0.96939564, 0.03060436, 0.93879128], requires_grad=True)

    """
    global __activated
    __activated = False  # pragma: no cover


def active_return():
    """Function that checks if the new return types system is activated.

    Returns:
        bool: Returns ``True`` if the new return types system is activated.

    **Example**

    By default, the new return types system is turned off:

    >>> active_return()
    False

    It can be activated:

    >>> enable_return()
    >>> active_return()
    True

    And it can also be deactivated:

    >>> enable_return()
    >>> active_return()
    True
    >>> disable_return()
    >>> active_return()
    False
    """
    return __activated



# Copyright 2018-2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains functions for adding the JAX interface
to a PennyLane Device class.
"""

# pylint: disable=too-many-arguments, no-member
import jax
import jax.numpy as jnp
import numpy as np
import semantic_version

import pennylane as qml
from pennylane.interfaces import InterfaceUnsupportedError
from pennylane.interfaces.jax import _raise_vector_valued_fwd
from pennylane.measurements import ProbabilityMP

dtype = jnp.float64


def _validate_jax_version():
    if semantic_version.match("<0.3.17", jax.__version__) or semantic_version.match(
        "<0.3.15", jax.lib.__version__
    ):
        msg = (
            "The JAX JIT interface of PennyLane requires version 0.3.17 or higher for JAX "
            "and 0.3.15 or higher JAX lib. Please upgrade these packages."
        )
        raise InterfaceUnsupportedError(msg)


def execute(tapes, device, execute_fn, gradient_fn, gradient_kwargs, _n=1, max_diff=1, mode=None):
    """Execute a batch of tapes with JAX parameters on a device.

    Args:
        tapes (Sequence[.QuantumTape]): batch of tapes to execute
        device (.Device): Device to use to execute the batch of tapes.
            If the device does not provide a ``batch_execute`` method,
            by default the tapes will be executed in serial.
        execute_fn (callable): The execution function used to execute the tapes
            during the forward pass. This function must return a tuple ``(results, jacobians)``.
            If ``jacobians`` is an empty list, then ``gradient_fn`` is used to
            compute the gradients during the backwards pass.
        gradient_kwargs (dict): dictionary of keyword arguments to pass when
            determining the gradients of tapes
        gradient_fn (callable): the gradient function to use to compute quantum gradients
        _n (int): a positive integer used to track nesting of derivatives, for example
            if the nth-order derivative is requested.
        max_diff (int): If ``gradient_fn`` is a gradient transform, this option specifies
            the maximum order of derivatives to support. Increasing this value allows
            for higher order derivatives to be extracted, at the cost of additional
            (classical) computational overhead during the backwards pass.
        mode (str): Whether the gradients should be computed on the forward
            pass (``forward``) or the backward pass (``backward``).

    Returns:
        list[list[float]]: A nested list of tape results. Each element in
        the returned list corresponds in order to the provided tapes.
    """
    # pylint: disable=unused-argument
    if max_diff > 1:
        raise InterfaceUnsupportedError("The JAX interface only supports first order derivatives.")

    for tape in tapes:
        # set the trainable parameters
        params = tape.get_parameters(trainable_only=False)
        tape.trainable_params = qml.math.get_trainable_indices(params)

    parameters = tuple(list(t.get_parameters()) for t in tapes)

    _validate_jax_version()

    if gradient_fn is None:
        return _execute_with_fwd(
            parameters,
            tapes=tapes,
            device=device,
            execute_fn=execute_fn,
            gradient_kwargs=gradient_kwargs,
            _n=_n,
        )

    return _execute(
        parameters,
        tapes=tapes,
        device=device,
        execute_fn=execute_fn,
        gradient_fn=gradient_fn,
        gradient_kwargs=gradient_kwargs,
        _n=_n,
    )


def _numeric_type_to_dtype(numeric_type):
    """Auxiliary function for converting from Python numeric types to JAX
    dtypes based on the precision defined for the interface."""

    single_precision = dtype is jnp.float32
    if numeric_type is int:
        return jnp.int32 if single_precision else jnp.int64

    if numeric_type is float:
        return jnp.float32 if single_precision else jnp.float64

    # numeric_type is complex
    return jnp.complex64 if single_precision else jnp.complex128


def _extract_shape_dtype_structs(tapes, device):
    """Auxiliary function for defining the jax.ShapeDtypeStruct objects given
    the tapes and the device.

    The host_callback.call function expects jax.ShapeDtypeStruct objects to
    describe the output of the function call.
    """
    shape_dtypes = []

    for t in tapes:
        shape = t.shape(device)

        tape_dtype = _numeric_type_to_dtype(t.numeric_type)
        shape_and_dtype = jax.ShapeDtypeStruct(tuple(shape), tape_dtype)

        shape_dtypes.append(shape_and_dtype)

    return shape_dtypes


def _execute(
    params,
    tapes=None,
    device=None,
    execute_fn=None,
    gradient_fn=None,
    gradient_kwargs=None,
    _n=1,
):  # pylint: disable=dangerous-default-value,unused-argument
    total_params = np.sum([len(p) for p in params])

    # Copy a given tape with operations and set parameters
    def cp_tape(t, a):
        tc = t.copy(copy_operations=True)
        tc.set_parameters(a)
        return tc

    @jax.custom_vjp
    def wrapped_exec(params):
        result_shapes_dtypes = _extract_shape_dtype_structs(tapes, device)

        def wrapper(p):
            """Compute the forward pass."""
            new_tapes = [cp_tape(t, a) for t, a in zip(tapes, p)]
            with qml.tape.Unwrap(*new_tapes):
                res, _ = execute_fn(new_tapes, **gradient_kwargs)

            # When executed under `jax.vmap` the `result_shapes_dtypes` will contain
            # the shape without the vmap dimensions, while the function here will be
            # executed with objects containing the vmap dimensions. So res[i].ndim
            # will have an extra dimension for every `jax.vmap` wrapping this execution.
            #
            # The execute_fn will return an object with shape `(n_observables, batches)`
            # but the default behaviour for `jax.pure_callback` is to add the extra
            # dimension at the beginning, so `(batches, n_observables)`. So in here
            # we detect with the heuristic above if we are executing under vmap and we
            # swap the order in that case.

            # pylint: disable=consider-using-enumerate
            for i in range(len(res)):
                if res[i].ndim > result_shapes_dtypes[i].ndim:
                    res[i] = res[i].T
            return res

        res = jax.pure_callback(wrapper, result_shapes_dtypes, params, vectorized=True)
        return res

    def wrapped_exec_fwd(params):
        return wrapped_exec(params), params

    def wrapped_exec_bwd(params, g):

        if isinstance(gradient_fn, qml.gradients.gradient_transform):
            for t in tapes:
                multi_probs = (
                    any(isinstance(m, ProbabilityMP) for m in t.measurements)
                    and len(t.measurements) > 1
                )

                if multi_probs:
                    raise InterfaceUnsupportedError(
                        "The JAX-JIT interface doesn't support differentiating QNodes that "
                        "return multiple probabilities."
                    )

            def non_diff_wrapper(args):
                """Compute the VJP in a non-differentiable manner."""
                p = args[:-1]
                dy = args[-1]

                new_tapes = [cp_tape(t, a) for t, a in zip(tapes, p)]
                vjp_tapes, processing_fn = qml.gradients.batch_vjp(
                    new_tapes,
                    dy,
                    gradient_fn,
                    reduction="append",
                    gradient_kwargs=gradient_kwargs,
                )

                partial_res = execute_fn(vjp_tapes)[0]
                res = processing_fn(partial_res)
                return np.concatenate(res)

            args = tuple(params) + (g,)
            vjps = jax.pure_callback(
                non_diff_wrapper,
                jax.ShapeDtypeStruct((total_params,), dtype),
                args,
            )

            param_idx = 0
            res = []

            # Group the vjps based on the parameters of the tapes
            for p in params:
                param_vjp = vjps[param_idx : param_idx + len(p)]
                res.append(param_vjp)
                param_idx += len(p)

            # Unwrap partial results into ndim=0 arrays to allow
            # differentiability with JAX
            # E.g.,
            # [Array([-0.9553365], dtype=float32), Array([0., 0.],
            # dtype=float32)]
            # is mapped to
            # [[Array(-0.9553365, dtype=float32)], [Array(0.,
            # dtype=float32), Array(0., dtype=float32)]].
            need_unstacking = any(r.ndim != 0 for r in res)
            if need_unstacking:
                res = [qml.math.unstack(x) for x in res]

            return (tuple(res),)

        def jacs_wrapper(p):
            """Compute the jacs"""
            new_tapes = [cp_tape(t, a) for t, a in zip(tapes, p)]
            with qml.tape.Unwrap(*new_tapes):
                jacs = gradient_fn(new_tapes, **gradient_kwargs)
            return jacs

        shapes = [
            jax.ShapeDtypeStruct((len(t.measurements), len(p)), dtype)
            for t, p in zip(tapes, params)
        ]
        jacs = jax.pure_callback(jacs_wrapper, shapes, params)
        vjps = [qml.gradients.compute_vjp(d, jac) for d, jac in zip(g, jacs)]
        res = [[jnp.array(p) for p in v] for v in vjps]
        return (tuple(res),)

    wrapped_exec.defvjp(wrapped_exec_fwd, wrapped_exec_bwd)
    return wrapped_exec(params)


# The execute function in forward mode
def _execute_with_fwd(
    params,
    tapes=None,
    device=None,
    execute_fn=None,
    gradient_kwargs=None,
    _n=1,
):  # pylint: disable=dangerous-default-value,unused-argument
    @jax.custom_vjp
    def wrapped_exec(params):
        def wrapper(p):
            """Compute the forward pass by returning the jacobian too."""
            new_tapes = []

            for t, a in zip(tapes, p):
                new_tapes.append(t.copy(copy_operations=True))
                new_tapes[-1].set_parameters(a)

            res, jacs = execute_fn(new_tapes, **gradient_kwargs)

            # On the forward execution return the jacobian too
            return res, jacs

        fwd_shape_dtype_struct = _extract_shape_dtype_structs(tapes, device)

        # params should be the parameters for each tape queried prior, assert
        # to double-check
        assert len(tapes) == len(params)

        jacobian_shape = []
        for t, p in zip(tapes, params):
            shape = t.shape(device) + (len(p),)
            _dtype = _numeric_type_to_dtype(t.numeric_type)
            shape = [shape] if isinstance(shape, int) else shape
            o = jax.ShapeDtypeStruct(tuple(shape), _dtype)
            jacobian_shape.append(o)

        res, jacs = jax.pure_callback(
            wrapper, tuple([fwd_shape_dtype_struct, jacobian_shape]), params
        )
        return res, jacs

    def wrapped_exec_fwd(params):
        res, jacs = wrapped_exec(params)
        return res, tuple([jacs, params])

    def wrapped_exec_bwd(params, g):

        _raise_vector_valued_fwd(tapes)

        # Use the jacobian that was computed on the forward pass
        jacs, params = params

        # Adjust the structure of how the jacobian is returned to match the
        # non-forward mode cases
        # E.g.,
        # [Array([[ 0.06695931,  0.01383095, -0.46500877]], dtype=float32)]
        # is mapped to
        # [[Array(0.06695931, dtype=float32), Array(0.01383095,
        # dtype=float32), Array(-0.46500877, dtype=float32)]]
        res_jacs = []
        for j in jacs:
            this_j = []
            for i in range(j.shape[1]):
                this_j.append(j[0, i])
            res_jacs.append(this_j)
        return tuple([tuple(res_jacs)])

    wrapped_exec.defvjp(wrapped_exec_fwd, wrapped_exec_bwd)
    res = wrapped_exec(params)

    tracing = any(isinstance(r, jax.interpreters.ad.JVPTracer) for r in res)

    # When there are no tracers (not differentiating), we have the result of
    # the forward pass and the jacobian, but only need the result of the
    # forward pass
    if len(res) == 2 and not tracing:
        res = res[0]

    return res



# Copyright 2018-2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module contains the functions needed for computing the spin observables.
"""
from pennylane import numpy as np

from .observable_hf import qubit_observable


def _spin2_matrix_elements(sz):
    r"""Builds the table of matrix elements
    :math:`\langle \bm{\alpha}, \bm{\beta} \vert \hat{s}_1 \cdot \hat{s}_2 \vert
    \bm{\gamma}, \bm{\delta} \rangle` of the two-particle spin operator
    :math:`\hat{s}_1 \cdot \hat{s}_2`.

    The matrix elements are evaluated using the expression

    .. math::

        \langle ~ (\alpha, s_{z_\alpha});~ (\beta, s_{z_\beta}) ~ \vert \hat{s}_1 &&
        \cdot \hat{s}_2 \vert ~ (\gamma, s_{z_\gamma}); ~ (\delta, s_{z_\gamma}) ~ \rangle =
        \delta_{\alpha,\delta} \delta_{\beta,\gamma} \\
        && \times \left( \frac{1}{2} \delta_{s_{z_\alpha}, s_{z_\delta}+1}
        \delta_{s_{z_\beta}, s_{z_\gamma}-1} + \frac{1}{2} \delta_{s_{z_\alpha}, s_{z_\delta}-1}
        \delta_{s_{z_\beta}, s_{z_\gamma}+1} + s_{z_\alpha} s_{z_\beta}
        \delta_{s_{z_\alpha}, s_{z_\delta}} \delta_{s_{z_\beta}, s_{z_\gamma}} \right),

    where :math:`\alpha` and :math:`s_{z_\alpha}` refer to the quantum numbers of the spatial
    function and the spin projection, respectively, of the single-particle state
    :math:`\vert \bm{\alpha} \rangle \equiv \vert \alpha, s_{z_\alpha} \rangle`.

    Args:
        sz (array[float]): spin-projection of the single-particle states

    Returns:
        array: NumPy array with the table of matrix elements. The first four columns
        contain the indices :math:`\bm{\alpha}`, :math:`\bm{\beta}`, :math:`\bm{\gamma}`,
        :math:`\bm{\delta}` and the fifth column stores the computed matrix element.

    **Example**

    >>> sz = np.array([0.5, -0.5])
    >>> print(_spin2_matrix_elements(sz))
    [[ 0.    0.    0.    0.    0.25]
     [ 0.    1.    1.    0.   -0.25]
     [ 1.    0.    0.    1.   -0.25]
     [ 1.    1.    1.    1.    0.25]
     [ 0.    1.    0.    1.    0.5 ]
     [ 1.    0.    1.    0.    0.5 ]]
    """

    n = np.arange(sz.size)

    alpha = n.reshape(-1, 1, 1, 1)
    beta = n.reshape(1, -1, 1, 1)
    gamma = n.reshape(1, 1, -1, 1)
    delta = n.reshape(1, 1, 1, -1)

    # we only care about indices satisfying the following boolean mask
    mask = np.logical_and(alpha // 2 == delta // 2, beta // 2 == gamma // 2)

    # diagonal elements
    diag_mask = np.logical_and(sz[alpha] == sz[delta], sz[beta] == sz[gamma])
    diag_indices = np.argwhere(np.logical_and(mask, diag_mask))
    diag_values = (sz[alpha] * sz[beta]).flatten()

    diag = np.vstack([diag_indices.T, diag_values]).T

    # off-diagonal elements
    m1 = np.logical_and(sz[alpha] == sz[delta] + 1, sz[beta] == sz[gamma] - 1)
    m2 = np.logical_and(sz[alpha] == sz[delta] - 1, sz[beta] == sz[gamma] + 1)

    off_diag_mask = np.logical_and(mask, np.logical_or(m1, m2))
    off_diag_indices = np.argwhere(off_diag_mask)
    off_diag_values = np.full([len(off_diag_indices)], 0.5)

    off_diag = np.vstack([off_diag_indices.T, off_diag_values]).T

    # combine the off diagonal and diagonal tables into a single table
    return np.vstack([diag, off_diag])


def spin2(electrons, orbitals):
    r"""Compute the total spin observable :math:`\hat{S}^2`.

    The total spin observable :math:`\hat{S}^2` is given by

    .. math::

        \hat{S}^2 = \frac{3}{4}N + \sum_{ \bm{\alpha}, \bm{\beta}, \bm{\gamma}, \bm{\delta} }
        \langle \bm{\alpha}, \bm{\beta} \vert \hat{s}_1 \cdot \hat{s}_2
        \vert \bm{\gamma}, \bm{\delta} \rangle ~
        \hat{c}_\bm{\alpha}^\dagger \hat{c}_\bm{\beta}^\dagger
        \hat{c}_\bm{\gamma} \hat{c}_\bm{\delta},

    where the two-particle matrix elements are computed as,

    .. math::

        \langle \bm{\alpha}, \bm{\beta} \vert \hat{s}_1 \cdot \hat{s}_2
        \vert \bm{\gamma}, \bm{\delta} \rangle = && \delta_{\alpha,\delta} \delta_{\beta,\gamma} \\
        && \times \left( \frac{1}{2} \delta_{s_{z_\alpha}, s_{z_\delta}+1}
        \delta_{s_{z_\beta}, s_{z_\gamma}-1} + \frac{1}{2} \delta_{s_{z_\alpha}, s_{z_\delta}-1}
        \delta_{s_{z_\beta}, s_{z_\gamma}+1} + s_{z_\alpha} s_{z_\beta}
        \delta_{s_{z_\alpha}, s_{z_\delta}} \delta_{s_{z_\beta}, s_{z_\gamma}} \right).

    In the equations above :math:`N` is the number of electrons, :math:`\alpha` refer to the
    quantum numbers of the spatial wave function and :math:`s_{z_\alpha}` is
    the spin projection of the single-particle state
    :math:`\vert \bm{\alpha} \rangle \equiv \vert \alpha, s_{z_\alpha} \rangle`.
    The operators :math:`\hat{c}^\dagger` and :math:`\hat{c}` are the particle creation
    and annihilation operators, respectively.

    Args:
        electrons (int): Number of electrons. If an active space is defined, this is
            the number of active electrons.
        orbitals (int): Number of *spin* orbitals. If an active space is defined,  this is
            the number of active spin-orbitals.

    Returns:
        pennylane.Hamiltonian: the total spin observable :math:`\hat{S}^2`

    Raises:
        ValueError: If electrons or orbitals is less than or equal to 0

    **Example**

    >>> electrons = 2
    >>> orbitals = 4
    >>> print(spin2(electrons, orbitals))
    (0.75) [I0]
    + (0.375) [Z1]
    + (-0.375) [Z0 Z1]
    + (0.125) [Z0 Z2]
    + (0.375) [Z0]
    + (-0.125) [Z0 Z3]
    + (-0.125) [Z1 Z2]
    + (0.125) [Z1 Z3]
    + (0.375) [Z2]
    + (0.375) [Z3]
    + (-0.375) [Z2 Z3]
    + (0.125) [Y0 X1 Y2 X3]
    + (0.125) [Y0 Y1 X2 X3]
    + (0.125) [Y0 Y1 Y2 Y3]
    + (-0.125) [Y0 X1 X2 Y3]
    + (-0.125) [X0 Y1 Y2 X3]
    + (0.125) [X0 X1 X2 X3]
    + (0.125) [X0 X1 Y2 Y3]
    + (0.125) [X0 Y1 X2 Y3]
    """

    if electrons <= 0:
        raise ValueError(f"'electrons' must be greater than 0; got for 'electrons' {electrons}")

    if orbitals <= 0:
        raise ValueError(f"'orbitals' must be greater than 0; got for 'orbitals' {orbitals}")

    sz = np.where(np.arange(orbitals) % 2 == 0, 0.5, -0.5)

    table = _spin2_matrix_elements(sz)

    s2_coeff = np.array([3 / 4 * electrons])
    s2_op = [[]]

    for i in table:
        s2_coeff = np.concatenate((s2_coeff, np.array([i[4]])))
        s2_op.append([int(i[0]), int(i[1]), int(i[2]), int(i[3])])

    return qubit_observable((s2_coeff, s2_op))


def spinz(orbitals):
    r"""Computes the total spin projection observable :math:`\hat{S}_z`.

    The total spin projection operator :math:`\hat{S}_z` is given by

    .. math::

        \hat{S}_z = \sum_{\alpha, \beta} \langle \alpha \vert \hat{s}_z \vert \beta \rangle
        ~ \hat{c}_\alpha^\dagger \hat{c}_\beta, ~~ \langle \alpha \vert \hat{s}_z
        \vert \beta \rangle = s_{z_\alpha} \delta_{\alpha,\beta},

    where :math:`s_{z_\alpha} = \pm 1/2` is the spin-projection of the single-particle state
    :math:`\vert \alpha \rangle`. The operators :math:`\hat{c}^\dagger` and :math:`\hat{c}`
    are the particle creation and annihilation operators, respectively.

    Args:
        orbitals (str): Number of *spin* orbitals. If an active space is defined, this is
            the number of active spin-orbitals.

    Returns:
        pennylane.Hamiltonian: the total spin projection observable :math:`\hat{S}_z`

    Raises:
        ValueError: If orbitals is less than or equal to 0

    **Example**

    >>> orbitals = 4
    >>> print(spinz(orbitals))
    (-0.25) [Z0]
    + (0.25) [Z1]
    + (-0.25) [Z2]
    + (0.25) [Z3]
    """

    if orbitals <= 0:
        raise ValueError(f"'orbitals' must be greater than 0; got for 'orbitals' {orbitals}")

    r = np.arange(orbitals)
    sz_orb = np.where(r % 2 == 0, 0.5, -0.5)
    table = np.vstack([r, r, sz_orb]).T

    sz_coeff = np.array([])
    sz_op = []

    for i in table:
        sz_coeff = np.concatenate((sz_coeff, np.array([i[2]])))
        sz_op.append([int(i[0]), int(i[1])])

    return qubit_observable((sz_coeff, sz_op))
